"""
    This script feeds MTtest object and compare the outcome with the reference
    implementation. 
    
    The tests covers:
        - various d
        - Usage of incremental API. 
        - Number of pois
        - Number of traces
"""
import pytest
from scalib.metrics import MTtest
import numpy as np


def reference(traces, x, d, pois):
    I0 = np.where(x == 0)[0]
    I1 = np.where(x == 1)[0]

    n0 = len(I0)
    n1 = len(I1)

    t0 = traces[I0]
    t1 = traces[I1]

    t0 = t0 - np.mean(t0, axis=0)
    t1 = t1 - np.mean(t1, axis=0)
    if d > 2:
        t0 = t0 / np.std(t0, axis=0)
        t1 = t1 / np.std(t1, axis=0)

    t0_mult = 1
    t1_mult = 1
    for d in range(d):
        t0_mult *= t0[:, pois[d, :]]
        t1_mult *= t1[:, pois[d, :]]

    print(t0_mult.shape)
    u0 = np.mean(t0_mult, axis=0)
    u1 = np.mean(t1_mult, axis=0)

    v0 = np.var(t0_mult, axis=0)
    v1 = np.var(t1_mult, axis=0)

    return (u0 - u1) / np.sqrt((v0 / n0) + (v1 / n1))


def test_ttest_d2_many_3():
    ns = 10237
    npois = 12342
    d = 2
    nc = 2
    n = 1 << 8

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    pois = np.random.randint(0, ns, (d, npois), dtype=np.uint32)
    traces += m[labels]

    t_ref = reference(traces, labels, d, pois)
    ttest = MTtest(d, pois)
    ttest.fit_u(traces, labels)
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d2_many_2():
    ns = 99
    npois = 12329
    d = 2
    nc = 2
    n = 1 << 8

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    pois = np.random.randint(0, ns, (d, npois), dtype=np.uint32)
    traces += m[labels]

    t_ref = reference(traces, labels, d, pois)
    ttest = MTtest(d, pois)
    ttest.fit_u(traces, labels)
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d2_many():
    ns = 10
    npois = 6
    d = 2
    nc = 2
    n = 1 << 17

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    pois = np.random.randint(0, ns, (d, npois), dtype=np.uint32)
    traces += m[labels]

    t_ref = reference(traces, labels, d, pois)
    ttest = MTtest(d, pois)
    ttest.fit_u(traces, labels)
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d2():
    ns = 100
    npois = 6302
    d = 2
    nc = 2
    n = 200

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    pois = np.random.randint(0, ns, (d, npois), dtype=np.uint32)
    traces += m[labels]

    t_ref = reference(traces, labels, d, pois)
    ttest = MTtest(d, pois)
    ttest.fit_u(traces, labels)
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d3():
    ns = 100
    npois = 10
    d = 3
    nc = 2
    n = 200

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    pois = np.random.randint(0, ns, (d, npois), dtype=np.uint32)
    traces += m[labels]

    t_ref = reference(traces, labels, d, pois)
    ttest = MTtest(d, pois)
    ttest.fit_u(traces, labels)
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d3_multiple_fit():
    ns = 1043
    d = 3
    nc = 2
    npois = 10021
    n = 40

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    pois = np.random.randint(0, ns, (d, npois), dtype=np.uint32)

    t_ref = reference(traces, labels, d, pois)
    ttest = MTtest(d, pois)
    for i in range(0, n, 10):
        ttest.fit_u(traces[i : i + 10, :], labels[i : i + 10])
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d3_multiple_fit():
    ns = 1043
    d = 3
    nc = 2
    npois = 10021
    n = 40

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    pois = np.random.randint(0, ns, (d, npois), dtype=np.uint32)

    t_ref = reference(traces, labels, d, pois)
    ttest = MTtest(d, pois)
    for i in range(0, n, 10):
        ttest.fit_u(traces[i : i + 10, :], labels[i : i + 10])
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)


def test_ttest_d2_multiple_fit():
    ns = 1043
    d = 2
    nc = 2
    npois = 10021
    n = 40

    m = np.random.randint(0, 2, (nc, ns))
    traces = np.random.randint(-4231, 4214, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    pois = np.random.randint(0, ns, (d, npois), dtype=np.uint32)

    t_ref = reference(traces, labels, d, pois)
    ttest = MTtest(d, pois)
    for i in range(0, n, 10):
        ttest.fit_u(traces[i : i + 10, :], labels[i : i + 10])
    t = ttest.get_ttest()
    assert np.allclose(t_ref, t, rtol=1e-3)



import pytest
import numpy as np
import queue
from scalib.metrics import SNR
from scalib.tools import ContextExecutor

n_traces = 10
ns = 5


def snr_inputs():
    snr = SNR(2**8, True)
    while (x := np_array_queue.get()) is not None:
        (traces, k) = x
        snr.fit_u(traces, k)


def get_traces():
    # Models a true acquisition process.
    k = np.zeros((n_traces, 16), dtype=np.uint16)
    traces = np.zeros((n_traces, ns), dtype=np.int16)
    return (traces, k)


def test_context_executor():
    snr = SNR(2**8, True)
    snr_workers = 1
    with ContextExecutor(max_workers=snr_workers) as e:
        # Initialize the future queue with dummy futures
        future_queue = [e.submit(lambda: None) for _ in range(snr_workers)]
        for _ in range(3):
            (traces, k) = get_traces()
            future_queue.append(e.submit(snr.fit_u, traces, k))
            # Wait on one future
            future_queue.pop(0).result()



import pytest
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA_sklearn
from scalib import ScalibError
from scalib.modeling import LDAClassifier, MultiLDA
import numpy as np
import scipy.stats
import pickle


def is_parallel(x, y):
    z = np.abs(x.dot(y))
    z2 = np.linalg.norm(x) * np.linalg.norm(y)
    return np.allclose(z, z2, rtol=1e-3)


def parallel_factor(x, y):
    """Return 1 if x and y are parallel and in the same direction, and -1 is
    they are in opposite directions.
    """
    x = x / np.linalg.norm(x)
    y = y / np.linalg.norm(y)
    return np.dot(x, y)


def test_lda_pickle():
    np.set_printoptions(threshold=np.inf)  # for debug
    ns = 10
    n_components = 2
    nc = 4
    n = 5000

    m = np.random.randint(0, 100, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    lda = LDAClassifier(nc, n_components)
    lda.fit_u(traces, labels, 1)
    lda.solve()

    dumped_lda = pickle.dumps(lda)
    lda = pickle.loads(dumped_lda)

    lda_ref = LDA_sklearn(solver="eigen", n_components=n_components)
    lda_ref.fit(traces, labels)

    lda_projection = lda.lda.get_projection()
    lda_ref_projection = lda_ref.scalings_[:, :n_components]
    projections_similar = all(
        is_parallel(x, y) for x, y in zip(lda_projection.T, lda_ref_projection.T)
    )
    print("lda_projection")
    print(lda_projection)
    print("lda_ref_projection")
    print(lda_ref_projection)
    assert projections_similar, (lda_projection, lda_ref_projection)
    # To correct if projection vectors are opposite.
    parallel_factors = np.array(
        [parallel_factor(x, y) for x, y in zip(lda_projection.T, lda_ref_projection.T)]
    )

    print("parallel_factors")
    print(parallel_factors)

    # generate means and cov in subspace
    traces_t = (lda_ref.scalings_[:, :n_components].T @ traces.T).T
    means_check = np.zeros((nc, n_components))
    for i in range(nc):
        I = np.where(labels == i)[0]
        means_check[i, :] = np.mean(traces_t[I, :], axis=0)
    traces_t = traces_t - means_check[labels, :]
    cov_check = np.cov(traces_t.T)

    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    prs = lda.predict_proba(traces)
    traces_t = (lda_ref.scalings_[:, :n_components].T @ traces.T).T
    prs_ref = np.zeros((len(traces), nc))
    for x in range(nc):
        prs_ref[:, x] = scipy.stats.multivariate_normal.pdf(
            traces_t, means_check[x], cov_check
        )
    prs_ref = (prs_ref.T / np.sum(prs_ref, axis=1)).T

    assert np.allclose(prs, prs_ref, rtol=1e-2)


def test_lda():
    np.set_printoptions(threshold=np.inf)  # for debug
    ns = 10
    n_components = 2
    nc = 4
    n = 500

    m = np.random.randint(0, 100, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    lda = LDAClassifier(nc, n_components)
    lda.fit_u(traces, labels, 1)
    lda.solve()

    lda_ref = LDA_sklearn(solver="eigen")
    lda_ref.fit(traces, labels)

    lda_projection = lda.lda.get_projection()
    lda_ref_projection = lda_ref.scalings_[:, :n_components]
    projections_similar = all(
        is_parallel(x, y) for x, y in zip(lda_projection.T, lda_ref_projection.T)
    )
    print("lda_projection")
    print(lda_projection)
    print("lda_ref_projection")
    print(lda_ref_projection)
    assert projections_similar, (lda_projection, lda_ref_projection)

    # Verify value of means by accessor
    ref_means = lda_ref.means_
    lda_means = lda.get_mus()
    assert np.allclose(ref_means, lda_means, rtol=1e-10)

    # Verify scatter matrix by accessor
    lda_scat = lda.get_sw() / n
    cov_ref = lda_ref.covariance_
    assert np.allclose(lda_scat, cov_ref, rtol=1e-5)

    # Not point of comparison with sklearn, but we here call
    # the accessor for the inter-class scatter matrix just to verify
    # that its worling.
    smat = lda.get_sb()

    # We can't do much more since sklearn has no way to reduce dimensionality of LDA.
    # e.g., comparing probas will fail


def test_lda_noproj():
    ns = 10
    n_components = 3
    nc = 4
    n = 500

    m = np.random.randint(0, 100, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    lda = LDAClassifier(nc, n_components)
    lda.fit_u(traces, labels, 1)
    lda.solve()

    lda_ref = LDA_sklearn(solver="eigen", n_components=n_components)
    lda_ref.fit(traces, labels)

    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    prs = lda.predict_proba(traces)
    prs_ref = lda_ref.predict_proba(traces)

    assert np.allclose(prs, prs_ref, rtol=1e-2, atol=1e-3)


def test_lda_fail_bad_traces():
    # Issue #56
    n = 100
    ns = 6
    nc = 4
    lda = LDAClassifier(nc, 3)
    traces_bad = np.random.randint(0, 1, (n, ns), dtype=np.int16)
    y = np.random.randint(0, nc, n, dtype=np.uint16)
    lda.fit_u(traces_bad, y, 0)
    with pytest.raises(ScalibError):
        lda.solve()


def test_multilda():
    x = np.random.randint(0, 256, (5000, 50), dtype=np.int16)
    y = np.random.randint(0, 256, (5000, 5), dtype=np.uint16)
    pois = [list(range(7 * i, 7 * i + 10)) for i in range(5)]
    lda = MultiLDA(5 * [256], 5 * [3], pois)
    lda.fit_u(x, y)
    lda.solve()
    x = np.random.randint(0, 256, (20, 50), dtype=np.int16)
    predicted_proba = lda.predict_proba(x)



from scalib.metrics import Ttest

from utils import gen_traces
import numpy as np


def main():
    # Parameters
    std = 2
    ntraces = 1000

    print(f"Fixed-vs-random TVLA on {ntraces} traces.")
    # Fix-vs-random TVLA (fixed-key vs random-key).
    # We generate new datasets, as the previous ones use random plaintext, and we needed fixed plaintext.
    traces_rk, _ = gen_traces(ntraces, std, random_key=True, random_plaintext=False)
    traces_fk, _ = gen_traces(ntraces, std, random_key=False, random_plaintext=False)
    # Ttest needs the lenght of the traces, and the test order (here, two)
    ttest = Ttest(2)
    # Ttest can be updated multiple times, and the order of the updates does not matter.
    ttest.fit_u(traces_rk, np.zeros((traces_rk.shape[0],), dtype=np.uint16))
    ttest.fit_u(traces_fk, np.ones((traces_fk.shape[0],), dtype=np.uint16))
    tt_result = ttest.get_ttest()
    print(
        "First-order t-statistic: ", ", ".join(f"{t: 5.01f}" for t in tt_result[0, :])
    )
    print(
        "Second-order t-statistic:", ", ".join(f"{t: 5.01f}" for t in tt_result[1, :])
    )


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.

# -- Project information -----------------------------------------------------

project = "SCALib"
copyright = "SIMPLE Contributors"


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinxcontrib.bibtex",
]
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

bibtex_bibfiles = ["refs.bib"]

# -- Options for HTML output -------------------------------------------------
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

autoclass_content = "both"

autosummary_ignore_module_all = False

autodoc_default_options = {
    "member-order": "bysource",
    "undoc-members": True,
    "special-members": False,
}

autodoc_typehints = "description"

# We do not need to install an build the package to generate the doc.
# We only have to mock dependencies, binary packages, and add the source to the
# python path.
autodoc_mock_imports = [
    "scalib._scalib_ext",
    "scalib.version",
    "scalib.build_config",
    "numpy",
    "cpuinfo",
]

linkcheck_ignore = [r'https://mathoverflow.net/a/273060']

import sys

sys.path.append("../src")

# from scalib.version import version
# The full version, including alpha/beta/rc tags
# release = version



from typing import Tuple

import numpy as np
import numpy.typing as npt

from scalib import _scalib_ext
from scalib.config import get_config
from scalib.modeling import RLDAClassifier


class RLDAInformationEstimator:
    r"""Amount of information that can be extracted from leakage with a RLDA model.

    This class can be used to compute Perceived Information (PI) or Training
    Information (TI) :footcite:p:`InfoBounds` bounds of a RLDA model
    :footcite:p:`RLDA` for a variable.

    It estimates the information using

    .. math::
        \hat{\mathrm{I}}(X,\mathbf{L}) = \mathrm{H}(X) + \frac{1}{|\mathcal{L}'|}
        \sum_{x \in \mathcal{X}} \sum_{\mathbf{l}\in \mathcal{L}'_x} \log_2 \hat{\mathsf{p}}[x|\mathbf{l}]

    PI or TI is obtained depending on the validation set used for the estimation.

    In order to obtain the PI : use a validation set independent of the training set of the model.

    In order to obtain the TI : use as validation set the full training set of the model.

    This object assumes that target variable is uniform
    and uses a clustered version of an RLDA model for faster computation,
    leading to bounds instead of exact values. If the bounds are not tight
    enough, change the parameters of the clustered model.

    The validation set is provided to this object by calling :meth:`fit_u`,
    and the estimated information is recovered with :meth:`get_information`.
    The estimator is further subject some variance due to the sampling of the
    validation set. The estimated standard deviation can be obtained with :meth:`get_deviation`.

    Examples
    --------

    >>> from scalib.modeling import RLDAClassifier
    >>> from scalib.metrics import RLDAInformationEstimator
    >>> import numpy as np
    >>> traces_model = np.random.randint(0,256,(5000,10),dtype=np.int16)
    >>> labels_model = np.random.randint(0,256,(5000,1),dtype=np.uint64)
    >>> rlda = RLDAClassifier(8, 3)
    >>> rlda.fit_u(traces_model, labels_model)
    >>> rlda.solve()
    >>> cl = rlda.get_clustered_model(0,0.5,100, False)
    >>> it = RLDAInformationEstimator(cl, 0)
    >>> traces_test = np.random.randint(0,256,(5000,10),dtype=np.int16)
    >>> labels_test = np.random.randint(0,256,(5000,),dtype=np.uint64)
    >>> it.fit_u(traces_test, labels_test)
    >>> pi_l,pi_u = it.get_information()

    See Also
    --------

    scalib.modeling.RLDAClassifier.get_clustered_model: Clustered model to use with ``RLDAInformationEstimator``.

    References
    ----------

    .. footbibliography::
    """

    def __init__(self, model: RLDAClassifier.ClusteredModel, max_popped_classes: int):
        r"""
        Parameters
        ----------
        model
            The clustered model, it is obtained by calling `get_clustered_model()` on an RLDA object.
        max_popped_classes
            Number of classes that are calculated without using the associated cluster. Corresponds to parameter :math:`\zeta` in :footcite:p:`RLDA`.
        """
        self._inner = _scalib_ext.ItEstimator(model._inner, max_popped_classes)

    def fit_u(self, traces: npt.NDArray[np.int16], x: npt.NDArray[np.uint64]):
        """Updates the estimatore with the given traces and the corresponding labels.

        This can be called multiple times with parts of the dataset: the state is accumulated.

        Parameters
        ----------
        traces : array_like, int16
            Array that contains the traces. Shape: ``(n,ns)``.
        x : array_like, uint64
            Label for each trace. Shape ``(n,)``.
        """
        assert traces.shape[0] == x.shape[0]
        self._inner.fit_u(traces, x, get_config())

    def get_information(
        self,
    ) -> Tuple[float, float]:
        """Returns the lower and upper bound information estimation."""
        return self._inner.get_information()

    def get_deviation(self) -> Tuple[float, float, int]:
        """Compute the approximate deviation of the information estimator for the lower and upper bound.
        Returns the deviation of the lower bound, the upper bound, and the number of traces used for the estimation.
        """
        raise self._inner.get_deviation()



"""SCALib runtime configuration.

The configuration is fetched from different sources, with the following priority.
The highest-priority is the local configuration enabled with the `Config`
context manager (which can be nested, and follows thread/async context).
If no such local configuration is activated, the default configuration is used.
This configuration has default startup values (which may depend on environment
variables) that can be changed with the `default_config` function.

Configurable Behaviors
----------------------

Progress bars
^^^^^^^^^^^^^

By default, a progress bar is shown for all SCALlib operations if the output
(stderr) is a terminal.
The printing of progress bars can be disabled with the `show_progress` argument of `Config`.

Thread pools
^^^^^^^^^^^^

All computationally-heavy operations of SCALib are run on a thread pool.
Multiple thread pools can be used simultaneously, although using only the
default thread pool should be enough for most applications.

The default thread pool can be configured through the default `Config` or using
value of the environment variable `SCALIB_NUM_THREADS`, e.g.:

.. code-block::

    SCALIB_NUM_THREADS=8 python3 XXX.py

If `SCALIB_NUM_THREADS` is not set, reasonable default for the typical use of
SCALib is taken (it is currently the one given `here
<https://docs.rs/num_cpus/1.13.1/num_cpus/fn.get_physical.html>`_), but might
not be optimal for your workload.

Thread pools can be used locally in place of the default thread pool by
enabling a local config using `Config` as a context manager.
When initializing a `Config`, a `ThreadPool` can be provided through the
`threadpool` argument, or a new `ThreadPool` is created if the `n_threads`
argument is provided.
If neither of these arguements is provided, the currently active thread pool is
used.

Example
--------

>>> from scalib.config import default_config, Config
>>> # Set the default ThreadPool to 10 threads and do not show progress bars.
>>> default_config(n_threads=10, show_progress=False)
>>> # As an exception, the following computations run on 5 threads, with progress bars.
>>> # (NB: enabling progress is needed as Config inherits the current context by default.)
>>> with Config(n_threads=5, show_progress=False).activate():
...     # Do some computations with SCALib...
...     pass
>>> # One can also re-use a Config.
>>> # This is convenient and also allows sending jobs from multiple python threads to a single thread pool.
>>> config = Config(n_threads=5)
>>> with config.activate():
...     pass
>>> with config.activate():
...     pass

Reference
---------

.. currentmodule:: scalib.config

.. autosummary::
   :toctree:
   :recursive:
   :nosignatures:

   Config
   default_config
   ThreadPool
"""

__all__ = ["default_config", "Config", "ThreadPool"]

import contextvars
import contextlib
from typing import Optional

from scalib import _scalib_ext

from .threading import ThreadPool, _default_num_threads


_current = contextvars.ContextVar("scalib config")


class Config:
    """SCALib configuration.

    Configuration applicable to all SCALib computations.
    """

    def __init__(
        self,
        *,
        threadpool: Optional[ThreadPool] = None,
        n_threads: Optional[int] = None,
        show_progress: Optional[bool] = None,
    ):
        self.inner = _scalib_ext.Config()
        if show_progress is not None:
            self.inner.show_progress(show_progress)
        else:
            self.inner.show_progress(get_config().inner.progress())
        if threadpool is not None:
            self.threadpool = threadpool
        elif n_threads is not None:
            self.threadpool = ThreadPool(n_threads)
        else:
            self.threadpool = get_config().threadpool

    @contextlib.contextmanager
    def activate(self):
        """Locally activate a Config."""
        restore_token = _current.set(self)
        try:
            yield
        finally:
            _current.reset(restore_token)


_default = Config(n_threads=_default_num_threads(), show_progress=True)
_current.set(_default)


def default_config(**kwargs):
    """Configure the default Config.

    Arguments are the same as Config.
    """
    # Use __init__ function instead of creating new object as we want to update
    # the object in-place.
    _default.__init__(**kwargs)


def get_config():
    """Get the current config, to give to Rust functions."""
    try:
        return _current.get()
    except LookupError:
        # TODO: use add_note on the orignal exception once we drop support for python 3.10.
        raise Exception(
            "Using SCALib from a thread without contextvars propagation is not supported.\n"
            + "We provide scalib.tools.ContextExecutor as a contextvars-aware replacement\n"
            + "for concurrent.futures.ThreadPoolExecutor.\n"
            + "See https://github.com/simple-crypto/SCALib/issues/89 for more information."
        )



import os
import numpy as np
import random

import pytest


@pytest.fixture(autouse=True)
def seed_random():
    init1 = random.SystemRandom().randrange(0, 2**32)
    init2 = random.SystemRandom().randrange(0, 2**32)
    print(f"Random seeds: random.seed({init1}); np.random.seed({init2})")
    random.seed(init1)
    np.random.seed(init2)



"""
Module only used by pytest.
"""

import pathlib
import runpy


def example_path(example_file):
    return str(pathlib.Path(__file__).parent / example_file)


def test_examples():
    for p in ["aes_attack.py", "aes_tvla.py", "aes_info.py"]:
        print(f"####### {p} #######")
        runpy.run_path(example_path(p), run_name="__main__")



import numpy as np
import numpy.typing as npt

from scalib import _scalib_ext
from scalib.config import get_config
import scalib.tools
import scalib.utils


class LDAClassifier:
    r"""Models the leakage :math:`\mathbf{l}` with :math:`n_s` dimensions using
    the linear discriminant analysis classifier (LDA) with integrated
    dimensionality reduction.

    Based on the training data, linear discriminant analysis build a linear
    dimentionality reduction to :math:`p` dimensions that maximizes class
    separation.
    Then, a multivariate gaussian template is fitted for each class (using the
    same covariance matrix for all the classes) in the reduced dimensionality
    space to predict leakage likelihood :footcite:p:`LDA`.

    Let :math:`\mathbf{W}` be the dimensionality reduction matrix of size
    (:math:`p`, :math:`n_s`). The likelihood is

    .. math::
            \mathsf{\hat{f}}(\mathbf{l} | x) =
                \frac{1}{\sqrt{(2\pi)^{p} \cdot |\mathbf{\Sigma} |}} \cdot
                \exp^{\frac{1}{2}
                    (\mathbf{W} \cdot \mathbf{l} - \mathbf{\mu}_x)
                    \mathbf{\Sigma}
                    ( \mathbf{W} \cdot \mathbf{l}-\mathbf{\mu}_x)'}

    where :math:`\mathbf{\mu}_x` is the mean of the leakage for class :math:`x` in
    the projected space (:math:`\mu_x = \mathbb{E}(\mathbf{W}\mathbf{l}_x)`, where
    :math:`\mathbf{l}_x` denotes the leakage traces of class :math:`x`) and
    :math:`\mathbf{\Sigma}` its covariance (:math:`\mathbf{\Sigma} =
    \mathbb{Cov}(\mathbf{W}\mathbf{l}_x - \mathbf{\mu}_x)`).

    :class:`LDAClassifier` provides the probability of each class with :meth:`predict_proba`
    thanks to Bayes' law such that

    .. math::
        \hat{\mathsf{pr}}(x|\mathbf{l}) = \frac{\hat{\mathsf{f}}(\mathbf{l}|x)}
                    {\sum_{x^*=0}^{n_c-1} \hat{\mathsf{f}}(\mathbf{l}|x^*)}.

    Example
    -------
    >>> from scalib.modeling import LDAClassifier
    >>> import numpy as np
    >>> # 5000 traces of length 10, with value between 0 and 255
    >>> traces = np.random.randint(0,256,(5000,10),dtype=np.int16)
    >>> # classes between 0 and 15
    >>> x = np.random.randint(0,16,5000,dtype=np.uint16)
    >>> lda = LDAClassifier(16,3)
    >>> lda.fit_u(traces, x)
    >>> lda.solve()
    >>> # predict classes for new traces
    >>> nt = np.random.randint(0,256,(20,10),dtype=np.int16)
    >>> predicted_proba = lda.predict_proba(nt)

    Notes
    -----
    This should have similar behavior as scikit-learn's `LDA
    <https://scikit-learn.org/stable/modules/lda_qda.html#lda-qda>`_, but it
    has better performance and numerical properties (at the cost of
    flexibility).

    References
    ----------

    .. footbibliography::

    Parameters
    ----------
    nc :
        Number of possible classes (e.g., 256 for 8-bit target). `nc` must
        be smaller than `2**16`.
    p :
        Number of dimensions in the linear subspace.
    """

    def __init__(self, nc: int, p: int):
        self.solved = False
        self.done = False
        self.p = p
        self._nc = nc
        self._ns = None
        self._init = False
        if p >= nc:
            raise ValueError("p must be at most nc")

    def fit_u(
        self,
        traces: npt.NDArray[np.int16],
        x: npt.NDArray[np.uint16],
        gemm_mode: int = 1,
    ):
        r"""Update statistical model estimates with fresh data.

        Parameters
        ----------
        traces :
            Array that contains the traces. The array must
            be of dimension `(n,ns)` and its type must be `int16`.
        x :
            Labels for each trace. Must be of shape `(n)` and
            must be `uint16`.
        gemm_mode:
            If 0: use matrixmultiply matrix multiplication.
            If n>0: use n threads with BLIS matrix multiplication.
            BLIS is only used on linux. Matrixmultiply is always used on other
            OSes.
            The BLIS threads (if > 1) do not belong to the SCALib threadpool.
        """
        traces = scalib.utils.clean_traces(traces, self._ns)
        x = scalib.utils.clean_labels(x, multi=False)
        if self.done:
            raise ValueError("Cannot fit_u after calling .solve(..., done=True).")
        if not self._init:
            self._init = True
            self._ns = traces.shape[1]
            self.acc = _scalib_ext.LdaAcc(self._nc, self._ns)
        # TODO maybe there is something smarter to do here w.r.t. number of
        # threads + investigate exact BLIS behavior.
        with scalib.utils.interruptible():
            self.acc.fit(traces, x, gemm_mode, get_config())
        self.solved = False

    def solve(self, done: bool = False):
        r"""Estimates the PDF parameters that is the projection matrix
        :math:`\mathbf{W}`, the means :math:`\mathbf{\mu}_x` and the covariance
        :math:`\mathbf{\Sigma}`.

        Parameters
        ----------
        done :
            True if the object will not be futher updated (clears some internal
            state, saving memory).

        Notes
        -----
        Once this has been called, predictions can be performed.
        """
        if not self._init:
            raise ValueError("Cannot .solve since .fit_u was never called.")
        if self.solved:
            raise ValueError(
                "Already called .solve() on this object, should not be called twice."
            )
        with scalib.utils.interruptible():
            self.lda = self.acc.lda(self.p, get_config())
        self.solved = True
        self.done = done

        if done:
            del self.acc

    def predict_proba(self, traces: npt.NDArray[np.int16]) -> npt.NDArray[np.float64]:
        r"""Computes the probability for each of the classes for the traces.

        Parameters
        ----------
        traces :
            Array that contains the traces. The array must be of dimension `(n,ns)`.

        Returns
        -------
        array_like, f64
            Probabilities. Shape `(n, nc)`.
        """
        if not self.solved:
            raise ValueError(
                "Call LDA.solve() before LDA.predict_proba() to compute the model."
            )
        with scalib.utils.interruptible():
            prs = self.lda.predict_proba(traces, get_config())
        return prs

    def __getstate__(self):
        dic = self.__dict__.copy()

        if "acc" in dic:
            dic["acc"] = dic["acc"].get_state()
        if "lda" in dic:
            dic["lda"] = dic["lda"].get_state()

        return dic

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)

        if "acc" in state:
            self.acc = _scalib_ext.LdaAcc.from_state(*state["acc"])
        if "lda" in state:
            self.lda = _scalib_ext.LDA.from_state(*state["lda"])

    def get_sw(self):
        r"""Return :math:`S_{W}` matrix (within-class scatter)."""
        return self.acc.get_sw()

    def get_sb(self):
        r"""Return :math:`S_{B}` matrix (between-class scatter)."""
        return self.acc.get_sb()

    def get_mus(self):
        r"""Return means matrix (classes means). Shape: ``(nc, ns)``."""
        return self.acc.get_mus()


class MultiLDA:
    """Perform LDA on `nv` distinct variables for the same leakage traces.

    While functionally similar to a simple for loop, this enables solving the
    LDA problems in parallel in a simple fashion. This also enable easy
    handling of Points Of Interest (POIs) in long traces.

    Parameters
    ----------
    ncs: array_like, int
        Number of classes for each variable. Shape `(nv,)`.
    ps: array_like, int
        Number of dimensions to keep after dimensionality reduction for each variable.
        Shape `(nv,)`.
    pois: list of array_like, int
        Indices of the POIs in the traces for each variable. That is, for
        variable ``i``, and training trace ``t``, ``t[pois[i]]`` is the input
        datapoints for the LDA.
    gemm_mode:
        See :func:`LDACLassifier.fit_u`.

    Examples
    --------
    >>> from scalib.modeling import MultiLDA
    >>> import numpy as np
    >>> # 5000 traces with 50 points each
    >>> traces = np.random.randint(0, 256, (5000,50),dtype=np.int16)
    >>> # 5 variables (8-bit), and 5000 traces
    >>> x = np.random.randint(0, 256, (5000, 5),dtype=np.uint16)
    >>> # 10 POIs for each of the 5 variables
    >>> pois = [list(range(7*i, 7*i+10)) for i in range(5)]
    >>> # Keep 3 dimensions after dimensionality reduction
    >>> lda = MultiLDA(5*[256], 5*[3], pois)
    >>> lda.fit_u(traces, x)
    >>> lda.solve()
    >>> # Predict the class for 20 traces.
    >>> nt = np.random.randint(0, 256, (20, 50), dtype=np.int16)
    >>> predicted_proba = lda.predict_proba(nt)
    """

    def __init__(self, ncs, ps, pois, gemm_mode: int = 1):
        self.pois = pois
        self.gemm_mode = gemm_mode
        self.ldas = [LDAClassifier(nc, p) for nc, p in zip(ncs, ps)]

    def fit_u(self, traces, x):
        """Update the LDA estimates with new training data.

        Parameters
        ----------
        traces : array_like, int16
            Array that contains the traces. The array must
            be of dimension `(n,ns)` and its type must be `int16`.
        x : array_like, uint16
            Labels for each trace. Must be of shape `(n, nv)` and
            must be `uint16`.
        """
        # Try to avoid the over-subscription of CPUs.
        num_threads = get_config().threadpool.n_threads
        if self.gemm_mode != 0:
            num_threads = max(1, num_threads // self.gemm_mode)
        with scalib.utils.interruptible():
            with scalib.tools.ContextExecutor(max_workers=num_threads) as executor:
                list(
                    executor.map(
                        lambda i: self.ldas[i].fit_u(
                            np.ascontiguousarray(traces[:, self.pois[i]]),
                            x[:, i],
                            self.gemm_mode,
                        ),
                        range(len(self.ldas)),
                    )
                )

    def solve(self, done: bool = False):
        """See `LDAClassifier.solve`."""
        # Put as much work as needed to fill rayon threadpool
        with scalib.utils.interruptible():
            with scalib.tools.ContextExecutor(
                max_workers=get_config().threadpool.n_threads
            ) as executor:
                list(executor.map(lambda lda: lda.solve(done), self.ldas))

    def predict_proba(self, traces):
        """Predict probabilities for all variables.

        See `LDAClassifier.predict_proba`.

        Parameters
        ----------
        traces : array_like, int16
            Array that contains the traces. The array must
            be of dimension `(n,ns)`.

        Returns
        -------
        list of array_like, f64
            Probabilities. `nv` arrays of shape `(n, nc)`.
        """
        # Put as much work as needed to fill rayon threadpool
        with scalib.utils.interruptible():
            with scalib.tools.ContextExecutor(
                max_workers=get_config().threadpool.n_threads
            ) as executor:
                return list(
                    executor.map(
                        lambda i: self.ldas[i].predict_proba(traces[:, self.pois[i]]),
                        range(len(self.ldas)),
                    )
                )



from scalib.modeling import LDAClassifier

from utils import gen_traces
import numpy as np


def main():
    nc = 256

    # Parameters
    std = 2
    ntraces_p = 10000
    ntraces_v = 1000

    print(
        "1. Generate simulated traces (Hamming weight + Gaussian noise) with parameters:"
    )
    print(f"    ntraces for profiling: {ntraces_p}")
    print(f"    ntraces for validation: {ntraces_v}")
    print(f"    noise std: {std}")
    print("")
    traces_p, labels_p = gen_traces(
        ntraces_p, std, random_key=True, random_plaintext=True
    )
    traces_v, labels_v = gen_traces(
        ntraces_v, std, random_key=True, random_plaintext=True
    )

    print("2. Profiling the value of the first Sbox.")
    print("")

    lda = LDAClassifier(nc=256, p=1)
    lda.fit_u(traces=traces_p, x=labels_p["x0"].astype(np.uint16))
    lda.solve()

    def eval_info(test_traces, test_labels):
        test_probas = lda.predict_proba(test_traces)
        test_probas = test_probas[range(test_probas.shape[0]), test_labels["x0"]]
        info = np.log2(nc) + np.mean(np.log2(test_probas))
        info_std = np.std(np.log2(test_probas)) / np.sqrt(test_traces.shape[0])
        return info, info_std

    print("3. PI: using validation traces")
    pi, pi_std = eval_info(traces_v, labels_v)
    print("PI [bits]", pi)
    print("PI stddev", pi_std)
    print("")

    print("4. TI: using profiling traces")
    ti, ti_std = eval_info(traces_p, labels_p)
    print("TI [bits]", ti)
    print("TI stddev", ti_std)


if __name__ == "__main__":
    main()



__all__ = [
    "preprocessing",
    "config",
    "attacks",
    "metrics",
    "modeling",
    "postprocessing",
    "tools",
    "ScalibError",
]

import cpuinfo

from .build_config import REQUIRE_AVX2
from .version import version as __version__

cpu = cpuinfo.get_cpu_info()

if REQUIRE_AVX2 and cpu["arch"] == "X86_64" and "avx2" not in cpu["flags"]:
    raise ImportError(
        "SCALib has been compiled with AVX2 instructions, which are not "
        + "supported by your CPU or OS. See "
        + "https://github.com/simple-crypto/SCALib/blob/main/README.rst "
        + "for compiling without AVX2 instructions."
    )

from scalib._scalib_ext import ScalibError



r"""
This module contains usefull preprocessing to manipulate side channel traces..

.. currentmodule:: scalib.preprocessing

.. autosummary::
   :toctree:
   :recursive:
   :nosignatures:

   QuantFitMethod
   Quantizer
"""
__all__ = ["QuantFitMethod", "Quantizer"]

from .quantization import QuantFitMethod, Quantizer



from scalib.metrics import SNR
from scalib.modeling import LDAClassifier
from scalib.attacks import FactorGraph, BPState
from scalib.postprocessing import rank_accuracy

from utils import sbox, gen_traces
import numpy as np


def main():
    nc = 256
    npoi = 2

    # Parameters
    std = 2
    ntraces_a = 40
    ntraces_p = 20000

    print(
        "1. Generate simulated traces (Hamming weight + Gaussian noise) with parameters:"
    )
    print(f"    ntraces for profiling: {ntraces_p}")
    print(f"    ntraces for attack:    {ntraces_a}")
    print(f"    noise std:             {std}")
    traces_p, labels_p = gen_traces(
        ntraces_p, std, random_key=True, random_plaintext=True
    )
    traces_a, labels_a = gen_traces(
        ntraces_a, std, random_key=False, random_plaintext=True
    )

    _, ns = traces_p.shape

    print("2. POI selection with SNR")
    print("    2.1 Compute SNR for 16 Sbox outputs xi")
    # y array with xi values
    x = np.zeros((ntraces_p, 16), dtype=np.uint16)
    for i in range(16):
        x[:, i] = labels_p[f"x{i}"]

    # estimate SNR
    snr = SNR(nc=nc)
    snr.fit_u(traces_p, x)
    snr_val = snr.get_snr()

    print("    2.2 Select POIs with highest SNR.")
    pois = [np.argsort(snr_val[i])[-npoi:] for i in range(16)]

    print("3. Profiling")
    # We build a LDA model (pooled Gaussian templates) for each of the 16
    # Sboxes (xi).

    print("    3.1 Build LDAClassifier for each xi")
    models = []
    for i in range(16):
        lda = LDAClassifier(nc=nc, p=1)
        lda.fit_u(
            traces=np.ascontiguousarray(traces_p[:, pois[i]]),
            x=labels_p[f"x{i}"].astype(np.uint16),
        )
        lda.solve()
        models.append(lda)

    print("    3.2 Get xi distributions from attack traces")
    probas = [models[i].predict_proba(traces_a[:, pois[i]]) for i in range(16)]

    print("4. Attack")
    print("    4.1 Create the SASCA Graph")
    graph_desc = f"""
                NC {nc}
                TABLE sbox   # The Sbox
                """

    for i in range(16):
        graph_desc += f"""
                VAR SINGLE k{i} # The key
                PUB MULTI p{i}  # The plaintext
                VAR MULTI x{i}  # Sbox output
                VAR MULTI y{i}  # Sbox input
                PROPERTY y{i} = k{i} ^ p{i} # Key addition
                PROPERTY x{i} = sbox[y{i}]  # Sbox lookup
                """

    # Initialize FactorGraph with the graph description and the required tables
    factor_graph = FactorGraph(graph_desc, {"sbox": sbox})

    print("    4.2 Create belief propagation state.")
    # We have to give the number of attack traces and the values for the public variables.
    bp = BPState(
        factor_graph,
        ntraces_a,
        {f"p{i}": labels_a[f"p{i}"].astype(np.uint32) for i in range(16)},
    )

    for i in range(16):
        bp.set_evidence(f"x{i}", probas[i])

    print("    4.3 Run belief propagation")
    for i in range(16):
        bp.bp_acyclic(f"k{i}")

    print("5. Attack evaluation")
    print("    5.1 Byte-wise attack")

    # correct secret key
    secret_key = []
    # distribution for each of the key bytes
    key_distribution = []
    # the best key guess of the adversary
    guess_key = []
    # rank for all the key bytes
    ranks = []

    for i in range(16):
        sk = labels_a[f"k{i}"][0]  # secret key byte
        distribution = bp.get_distribution(f"k{i}")

        guess_key.append(np.argmax(distribution))
        ranks.append(256 - (np.argsort(distribution) == sk).nonzero()[0][0])

        secret_key.append(sk)
        key_distribution.append(distribution)

    print("")
    print("        secret key (hex):", " ".join(["%3x" % (x) for x in secret_key]))
    print("        best key   (hex):", " ".join(["%3x" % (x) for x in guess_key]))
    print("        key byte ranks  :", " ".join(["%3d" % (x) for x in ranks]))
    print("")

    print("   5.2 Estimate full log2 key rank:")
    key_distribution = np.array(key_distribution)

    # Scores are negative log-likelihoods.
    # Put a lower-bound on the probabilities, otherwise we might get NaNs.
    scores = -np.log2(np.maximum(key_distribution, 2**-128))
    rmin, r, rmax = rank_accuracy(scores, secret_key, 1)

    lrmin, lr, lrmax = (np.log2(rmin), np.log2(r), np.log2(rmax))
    print("")
    print(f"        {lrmin} < {lr} < {lrmax}")
    print("")


if __name__ == "__main__":
    print("attack")
    main()



r"""
The Welch's :math:`t`-test can be used to highlight a difference in the means
of two distributions. To do so, a `t` statistic is derived following the
expression:

.. math::
    t = \frac{\mu_0 - \mu_1}{\sqrt{\frac{v_0}{n_0} + \frac{v_1}{n_1}}}

where :math:`\mu_0` (resp. :math:`\mu_1`) is the estimated moment of the first
(resp.second) population and :math:`\frac{v_0}{n_0}` the variance of its
estimate from :math:`n_0` samples. In the context of side-channel analysis, many of
these statistical tests are performed independently for each point of the traces. 
See :footcite:p:`TVLA` for additional details.

In this module, the definition of :math:`\mu` and :math:`v` are adapted to perform
higher-order univariate and multivariate :math:`t`-test to compare higher-order moments of
two distributions:

- *Higher-order t-tests* pre-process the populations by elevating the traces at
  the :math:`d`-th power, after subtracting the mean (for :math:`d>1`) and
  normalizing the variance (for :math:`d>2`).
- *Multivariate t-test* are a variant of higher-order t-tests where a product
  of :math:`d` samples of the trace is use instead of the :math:`d`-th power of
  a single sample. Mean and variance normalization are applied similarly to
  higher-order t-tests.

.. currentmodule:: scalib.metrics

.. autosummary::
    :toctree:
    :nosignatures:
    :recursive:

    Ttest
    MTtest

Warning
^^^^^^^

Ttest should not be used alone as a standalone evaluation tool because of its
qualitative nature. See :footcite:p:`how_not_ttest,critical_analysis_iso17825`
for cautionary notes.


Implementations Details
^^^^^^^^^^^^^^^^^^^^^^^

In order to enable both efficient one-core and parallelized performance of the
:math:`t`-test implementation, SCALib uses the one-pass formula for estimation
arbitrary order statistical moment from :footcite:p:`onepass_moments` and its
application to side-channel context in :footcite:p:`TVLA`.

Concretely, the implementations first compute an estimation of the required
statistical moments using a two-passes algorithms (first pass to compute the
mean and the variances, and a second pass to compute the centered products).
This new estimation is then used to update the current estimation using the
merging rule from :footcite:p:`onepass_moments`. To enable multi-threading,
SCALib internally divides the fresh traces into smaller independent chunks and
then merges the output of each threads using :footcite:p:`onepass_moments`. 

As a conclusion, the performance of the SCALib improves if the two-passes
algorithm can be used on large chunks. Hence, it is recommended to feed a large
enough amount of data for every call to `fit_u()`. 

References
^^^^^^^^^^

.. footbibliography::
"""

import numpy as np
import numpy.typing as npt

from scalib import _scalib_ext
from scalib.config import get_config
import scalib.utils


class Ttest:
    r"""Univariate (higher order) :math:`t`-test.

    Concretely, :math:`\mu[j]`'s and :math:`v[j]`'s are computed for
    the `d`-th statistical moment to test at all the indexes (`j`) in the
    traces. These are derived based on all the provided traces :math:`l[:,j]`, that
    can be provided through multiple calls to `fit_u`.

    The statistic `t[j]` is then derived as:

    - for `d=1`:

    .. math::
        \mu[j] = \frac{1}{n} \sum_{i=0}^{n-1} l[i,j]

    - for `d=2`:

    .. math::
        \mu[j] = \frac{1}{n} \sum_{i=0}^{n-1} (l[i,j] - \bar{l}[:,j])^2

    - for `d>2`:

    .. math::
        \mu[j] = \frac{1}{n} \sum_{i=0}^{n-1} \left(\frac{l[i,j] - \bar{l}[:,j])}{\sigma_{l[:,j]}}\right)^d

    where, :math:`\bar{l}` denotes the estimated mean of `l` and
    :math:`\sigma_l` its standard deviation.

    A :math:`d`-th order t-test automatically includes the computation of all the lower order t-tests.

    Example
    --------
    >>> from scalib.metrics import Ttest
    >>> import numpy as np
    >>> traces = np.random.randint(0,256,(100,200),dtype=np.int16)
    >>> x = np.random.randint(0,2,100,dtype=np.uint16)
    >>> ttest = Ttest(d=3)
    >>> ttest.fit_u(traces,x)
    >>> t = ttest.get_ttest()

    Parameters
    ----------
    d : int
        Maximal statistical order of the :math:`t`-test.
    """

    def __init__(self, d: int):
        self._d = d
        self._ns = None
        self._init = False

    def fit_u(self, traces: npt.NDArray[np.int16], x: npt.NDArray[np.uint16]):
        r"""Updates the Ttest estimation with new data.

        This method may be called multiple times.

        Parameters
        ----------
        traces :
            Array that contains the traces. The array must be of dimension
            `(n, ns)`.
        x :
            Set in which each trace belongs. Must be of shape `(n,)` and must
            contain only `0` and `1`.
        """
        traces = scalib.utils.clean_traces(traces, self._ns)
        x = scalib.utils.clean_labels(x, multi=False)
        if not self._init:
            self._init = True
            self._ns = traces.shape[1]
            self._ttest = _scalib_ext.Ttest(self._ns, self._d)
        if traces.shape[0] != x.shape[0]:
            raise ValueError(
                f"Number of traces {traces.shape[0]} does not match size of classes array {x.shape[0]}."
            )
        with scalib.utils.interruptible():
            self._ttest.update(traces, x, get_config())

    def get_ttest(self) -> npt.NDArray[np.float64]:
        r"""Return the current Ttest estimation with an array of shape `(d,ns)`."""
        if not self._init:
            raise ValueError("Need to call .fit_u at least once.")
        with scalib.utils.interruptible():
            return self._ttest.get_ttest(get_config())


class MTtest:
    r"""Multivariate :math:`t`-test.

    Concretely, :math:`\mu[j]`'s and :math:`v[j]`'s are computed to
    use in the :math:`t`-test definition. Especially, :math:`\mu[j]`'s is
    derived such as:

    For `d=2`:

    .. math::
        \mu[j] = \frac{1}{n} \sum_{i=0}^{n-1} \prod_{d'=0}^{1}l[i,pois[d',j]] - \bar{l}[:,pois[d',j]]

    For `d>2`:

    .. math::
        \mu[j] = \frac{1}{n} \sum_{i=0}^{n-1} \prod_{d'=0}^{d-1}
        \frac{l[i,pois[d',j]] -
        \bar{l}[:,pois[d',j]]}{\sigma_{l[:,pois[d',j]]}}

    Especially, :math:`\bar{l}` denotes the estimated mean of `l` and
    :math:`\sigma_l` its standard deviation. `pois` defines the points in the
    traces for which the (normalized) product will be tested.

    Parameters
    ----------
    d : int
        Number of variables of the :math:`t`-test.
    pois : array_like, uint32
        Array of shape ``(d,n_pois)``. Each column in `pois` will result in a
        :math:`t`-test of the product of the traces at these :math:`d` indexes
        (each index is between 0 and ns-1).
        If an index is repeated in a column, the corresponding test uses a
        higher-order moment for the corresponding point in the trace.

    Examples
    --------
    >>> from scalib.metrics import MTtest
    >>> import numpy as np
    >>> traces = np.random.randint(0,256,(100,200),dtype=np.int16)
    >>> # Take as POIs each point in the trace combined with any of the 10 following samples.
    >>> pois = np.array([[x, x+d] for x in range(200) for d in range(10) if x + d < 200], dtype=np.uint32).T
    >>> x = np.random.randint(0,2,100,dtype=np.uint16)
    >>> mttest = MTtest(d=2,pois=pois)
    >>> mttest.fit_u(traces,x)
    >>> t = mttest.get_ttest()

    """

    def __init__(self, d, pois):
        self._pois = pois
        self._d = d
        self._ns = len(pois[0, :])

        self._mttest = _scalib_ext.MTtest(d, self._pois)

    def fit_u(self, traces: npt.NDArray[np.int16], x: npt.NDArray[np.uint16]):
        r"""Updates the MTtest estimation with new traces.
        This method may be called multiple times.

        Parameters
        ----------
        traces :
            Array that contains the signal. The array must
            be of dimension `(n, ns)`.
        x :
            Set in which each trace belongs. Must be of shape `(n,)`
            and must contain only `0` and `1`.
        """
        traces = scalib.utils.clean_traces(traces)
        x = scalib.utils.clean_labels(x, multi=False)
        if x.shape[0] != traces.shape[0]:
            raise ValueError(
                f"Number of traces {traces.shape[0]} does not match size of classes array {x.shape[0]}."
            )
        with scalib.utils.interruptible():
            self._mttest.update(traces, x, get_config())

    def get_ttest(self) -> npt.NDArray[np.float64]:
        r"""Return the current MTtest estimation with an array of shape
        `(n_pois,)`. Each element in that array corresponds to a test defined by
        `pois`.

        """
        with scalib.utils.interruptible():
            return self._mttest.get_ttest(get_config())



import pytest
from scalib.preprocessing import Quantizer, QuantFitMethod
import numpy as np


def test_quantizer():
    ns = 200
    fitting_traces = np.random.randn(500, ns)
    traces = np.random.randn(5000, ns)

    quantizer = Quantizer.fit(fitting_traces, QuantFitMethod.moment())
    quantized_traces = quantizer.quantize(traces)

    quantized_traces = quantizer.quantize(8 * traces, True)
    with pytest.raises(ValueError):
        quantized_traces = quantizer.quantize(8 * traces)

    quantizer = Quantizer.fit(fitting_traces, QuantFitMethod.bounds(4.0))
    quantized_traces = quantizer.quantize(traces)

    reconstruction = (quantized_traces / quantizer._scale + quantizer._shift).astype(
        np.float64
    )
    reconstruction_error: np.float64 = np.linalg.norm(traces - reconstruction, axis=1)
    assert (reconstruction_error <= 10**-2).all()



r"""Estimation of the rank of a true key given likelihood of sub-keys.

Rank estimation estimates the rank of the full (e.g. 128-bit) key based on a
score for each value of its (e.g 8-bit) sub-keys. The scores must be additive
and positive, with lower score meaning lower rank (e.g. negative
log-likelihoods).

The `rank_accuracy` function allows to specify the desired precision of the
bound in bits. It is the high-level, and its usage is recommended over the
lower-level `rank_nbin` function.
That function gives direct, lower-level control to the core algorithm, allowing
to specify the number of bins in the histograms (whereas `rank_accuracy` tunes
this parameter automatically).

Examples
--------

>>> from scalib.postprocessing import rank_accuracy
>>> import numpy as np
>>> # define the correct key
>>> key = np.random.randint(0,256,16)
>>> # Derive the score for each key byte
>>> scores = np.ones((16,256)) * 1E-5
>>> scores[np.arange(16),key] = 1
>>> # Compute the full key rank (correct key must have rank 1).
>>> (rmin,r,rmax) = rank_accuracy(-np.log(scores),key)
>>> assert r == 1


Reference
---------

.. currentmodule:: scalib.postprocessing.rankestimation

.. autosummary::
    :toctree:
    :nosignatures:
    :recursive:

    rank_accuracy
    rank_nbin

Notes
-----
The rank estimation algorithm is based on :footcite:p:`rankest_histograms`,
with the following
optimization: computation of histogram bins with higher score than the expected
key is skipped, since it has no impact on the final rank.

References
----------

.. footbibliography::
"""

import math

from scalib import _scalib_ext
from scalib.config import get_config
import scalib.utils


def rank_nbin(costs, key, nbins, method="hist"):
    r"""Estimate the rank of the full keys based on scores based on histograms.

    Warning: this is a low-level function, you probably want to use
    `rank_accuracy` instead.

    Parameters
    ----------
    costs : array_like, f64
        Cost for each of the sub-keys. Array must be of shape `(ns,nc)` where
        `ns` is the number of sub-keys, `nc` the possible values of each
        sub-keys.
    key : array_like, int
        Correct full key split in sub-keys. Array must be of shape `(ns,)`.
    nbins : int
        Number of bins for each of the distributions.
    method : string
        Method used to estimate the rank. Can be the following:

        * "hist": using histograms (default).
        * "naive": enumerate possible keys (very slow).
        * "histbignum": using NTL library, allows better precision.

    Returns
    -------
    (rmin, r, rmax): (float, float, float)

            - **rmin** is a lower bound for the key rank.
            - **r** is the estimated key rank.
            - **rmax** is an upper bound for the key rank.
    """
    with scalib.utils.interruptible():
        return _scalib_ext.rank_nbin(
            costs, key, nbins, _choose_merge_value(costs), method, get_config()
        )


def rank_accuracy(costs, key, acc_bit=1.0, method="hist", max_nb_bin=2**26):
    r"""Estimate the rank of the full keys based on scores based on histograms.

    Parameters
    ----------
    costs : array_like, f64
        Cost for each of the sub-keys. Array must be of shape `(ns,nc)` where
        `ns` is the number of sub-keys, `nc` the possible values of each
        sub-keys.
    key : array_like, int
        Correct full key split in sub-keys. Array must be of shape `(ns,)`.
    acc_bit : f64, default: 1.0
        Expected log2 accuracy for the key rank estimation.

        The algorithms attempts to get a result such that `rmax/rmin < 2^acc_bit`,
        but may not achieve this if the result is computationally intractable.
        In such a case, rank estimate and rank bounds are still returned, but
        the inequality may be violated.
    method : string
        Method used to estimate the rank. Can be the following:

        * "hist": using histograms (default).
        * "ntl": using NTL library, allows better precision.
    max_nb_bin : int, default: 2**26
        Maximum number of bins to use.
        This fixes an upper bound on the computational cost of the algorithm
        (if too low, the requested accuracy might not be reached).

    Returns
    -------
    (rmin, r, rmax): (float, float, float)

            - **rmin** is a lower bound for the key rank.
            - **r** is the estimated key rank.
            - **rmax** is an upper bound for the key rank.
    """
    with scalib.utils.interruptible():
        return _scalib_ext.rank_accuracy(
            costs,
            key,
            2.0**acc_bit,
            _choose_merge_value(costs),
            method,
            max_nb_bin,
            get_config(),
        )


def _choose_merge_value(costs):
    """The merge parameter is the number of sub-keys to merge in a
    brute-force manner before computing histograms. Merging may improve
    accuracy at the expense of running time.
    Here we limit sub-histograms to 2**16 values.
    """
    return min(len(costs), max(1, int(16 / math.log2(max(len(c) for c in costs)))))



from typing import Sequence, Mapping, Union, Optional
from dataclasses import dataclass
from enum import IntEnum
import numpy as np
import numpy.typing as npt

from scalib import _scalib_ext
from scalib.config import get_config

__all__ = ["FactorGraph", "BPState", "GenFactor"]


@dataclass
class GenFactor:
    r"""Generic factor which can implement arbitrary functions between
    variables. A factor can be declared as `MULTI` if the factor changes between
    executions. For most use cases, where the implementation of the factor is
    static, `SINGLE` is sufficient. An example of a "butterfly" factor which
    computes the sum and difference of the inputs is shown below.

    >>> # Describe and generate the SASCAGraph
    >>> graph_desc = '''
    ... NC 13 # arbitrary prime number
    ... VAR MULTI a # inputs
    ... VAR MULTI b
    ... VAR MULTI c # outputs
    ... VAR MULTI d
    ... GENERIC SINGLE f # declare the factor
    ... PROPERTY F0: f(a,b,c,d) # declare the relationship c = a + b and d = a - b'''
    >>> factor = []
    >>> nc = 13
    >>> # Construct the factor table
    >>> for a in range(nc):
    ...     for b in range(nc):
    ...         factor.append([a, b, (a + b) % nc, (a - b) % nc])
    >>> factor = GenFactor.sparse_functional(np.array(factor, dtype=np.uint32))
    >>> graph = FactorGraph(graph_desc)
    >>> bpstate = BPState(graph, nexec=1, gen_factors={"f": factor})

    Notes
    -----
    In general, the computational cost of sparse_functional factors is linear
    in the number of rows. For dense factors, the cost is exponential in the
    number of axes. Therefore, one should prefer to use built-in functions (^, &, !, etc)
    whenever possible to minimize the performance penalty of generic factors,
    and to keep the number entries in the factor small when they must be used.
    """

    class GenFactorKind(IntEnum):
        """Choices of generic factor types"""

        DENSE = 0
        SPARSE_FUNCTIONAL = 1

    kind: GenFactorKind
    factor: npt.NDArray

    @classmethod
    def dense(cls, factor: npt.NDArray[np.float64]):
        """A dense factor.

        ``factor`` is a n-dimensional array, each axis corresponds to one
        variable, entries in the array are probabilities.
        """
        assert len(set(factor.shape)) == 1
        assert factor.dtype == np.float64
        return cls(cls.GenFactorKind.DENSE, factor)

    @classmethod
    def sparse_functional(cls, factor: npt.NDArray[np.uint32]):
        """A sparse functional factor.

        ``factor`` is a 2D array, each row corresponding to an entry in the
        factor, and in each row, the values are the values of the variables.
        Each row corresponds to setting the factor entry to 1. Omitting a factor
        entry implicitly assigns that entry to 0. Sparse factors leads to better
        performance and lower memory usage if the number of rows is much smaller
        than the total number of entries in the factor.
        """
        assert len(factor.shape) == 2
        assert factor.dtype == np.uint32
        return cls(cls.GenFactorKind.SPARSE_FUNCTIONAL, factor)


CstValue = Union[int, Sequence[int]]
ValsAssign = Mapping[str, CstValue]
GenFactors = Mapping[str, Union[GenFactor, Sequence[GenFactor]]]


class FactorGraph:
    r"""FactorGraph allows to run Soft Analytical Side-Channel Attacks (SASCA).

    A SASCA is based on a set of variables, on knowledge of
    relationships between those variables and information about the values of
    the variables.

    Variables have values in a binary finite field of size `nc`, and are viewed
    as bit vectors. Those values are represented as integers in [0, nc).
    Variables can be qualified as `SINGLE` or `MULTI`, which relates
    to the multiple-execution feature of `FactorGraph`: when performing a
    SASCA, it is useful to acquire multiple execution traces. In these
    executions, some variables stay the same (e.g. an encryption key), and other
    change (e.g. plaintext, masked variables, etc.).
    The `SINGLE` qualifier should be used for variables that remain the same
    and `MULTI` for variables that change.
    `FactorGraph` will then build a graph where each `MULTI` variable is
    replicated `n` times (once for each execution), as well as all the
    relationships that relate at least on `MULTI` variable.

    Relationships between variables are bitwise XOR, bitwise AND, bitwise OR,
    bitiwise negation, modular addition, modular multiplication and lookup
    table. A lookup table can describe any function that maps a single variable
    to another variable.
    Description of `nc`, the variables, and the relationships is given in a
    text format specified below.

    Finally, variables are of two kinds: the public values and the variables.
    The public values have a single known value, while variables have an
    uncertain value modelled as a probability distribution.
    The SASCA computes these distributions from the relationships encoded by
    the graph, and from prior information.
    The prior distributions are by default uniform, but this can be changed.

    An attack attempting to recover the secret key byte `k` is shown below.

    >>> # Describe and generate the SASCAGraph
    >>> graph_desc = '''
    ...     # Small unprotected Sbox example
    ...     NC 256 # Graph over GF(256)
    ...     TABLE sbox   # Sbox
    ...     VAR SINGLE k # key (to recover !)
    ...     PUB MULTI p  # plaintext (known)
    ...     VAR MULTI x # Sbox input
    ...     VAR MULTI y # Sbox output (whose leakage is targeted)
    ...     PROPERTY x = k ^ p   # Key addition
    ...     PROPERTY y = sbox[x] # Sbox lookup
    ...     '''
    >>> sbox = np.arange(256, dtype=np.uint32) # don't use this S-box ;)
    >>> graph = FactorGraph(graph_desc, {"sbox": sbox})
    >>> # n is the number of traces for our attack.
    >>> n = 2
    >>> plaintexts = np.array([0, 1], dtype=np.uint32)
    >>> bp = BPState(graph, n, {"p": plaintexts})
    >>> y_leakage = np.random.rand(2, 256) # this might come from an LDA
    >>> y_leakage = y_leakage / y_leakage.sum(axis=1, keepdims=True)
    >>> bp.set_evidence("y",y_leakage)
    >>> # Solve graph
    >>> bp.bp_loopy(it=3, initialize_states=True)
    >>> # Get key distribution and derive key guess
    >>> k_distri = bp.get_distribution("k")
    >>> key_guess = np.argmax(k_distri)

    By running a belief propagation algorithm (see [1]_), the distributions on all
    the variables are updated based on their initial distributions. The
    `SASCAGraph` can be solved by using `run_bp()`.

    Notes
    -----

    **The graph description format** is a text line-oriented format (each line
    of text is a statement) that describes a set of variables and relationships
    between those variables.
    The ordering of the statements is irrelevant and whitespace is irrelevant
    except for newlines and around keywords. End-of-line comments start with
    the `#` symbol.
    The statements are:

    - `NC <nc>`: specifies the field size (must be a power of two). There must
      be one `NC` statement in the description.
    - `PUB SINGLE|MULTI variable_name`: declares a public variable.  `variable_name`
      is an identifier of the variable (allowed characters are letters, digits
      and underscore). One of the qualifiers `SINGLE` or `MULTI` must be given.
    - `VAR SINGLE|MULTI variable_name`: declares a variables.
    - `PROPERTY w = x ^ y ^ z`: declares a bitwise XOR property. There can be any
      number of operands.
    - `PROPERTY z = x & y`: declares a bitwise AND property.
    - `PROPERTY x = t[y]`: declares a LOOKUP property (`y` is the lookup of the
      table `t` at index `y`). No public variable is allowed in this property.
    - `PROPERTY x = !y`: declares a bitwise NOT property.
      No public variable is allowed in this property.
    - `PROPERTY x = y + z - w`: declares a modular sum (+ and - are supported,
      with arbitrary number of terms).
    - `PROPERTY f(x, y, z)`: declares a "Generic factor" property, f must be
       declared as a GENERIC.
    - `TABLE` t = [0, 3, 2, 1]`: Declares a table that can be used in a LOOKUP.
      The values provided in the table must belong to the interval [0, nc).
      The initialization expression can be omitted from the graph description
      (e.g. `TABLE t`) and be given with `tables` parameter.
    - `GENERIC SINGLE|MULTI f`: declares a "Generic factor" f. A generic factor
        can be used to define an arbitrary functional relationships between
        variables. See the docs for ``GenFactor`` for more information and examples.


    **Note**: if the `MULTI` feature doesn't match your use-case, using only
    `SINGLE` variables works (or set the number of execution to 1).


    .. [1] "Soft Analytical Side-Channel Attacks". N. Veyrat-Charvillon, B.
       Gérard, F.-X. Standaert, ASIACRYPT2014.

    Parameters
    ----------
    graph: string
        The graph description.
    """

    def __init__(
        self,
        graph_text: str,
        tables: Optional[Mapping[str, npt.NDArray[np.uint32]]] = None,
    ):
        if tables is None:
            tables = dict()
        self._inner = _scalib_ext.FactorGraph(graph_text, tables)

    def sanity_check(
        self,
        pub_assignment: ValsAssign,
        var_assignment: ValsAssign,
        factor_assignment: Optional[GenFactors] = None,
    ):
        """Verify that the graph is compatible with example variable assignments.

        If the graph is not compatible, raise a ``ValueError``.

        Parameters
        ----------
        pub_assignment:
            For each public variable its value for all test executions.
        var_assignment:
            For each non-public variable its value for all test executions.
        factor_assignment:
            The probability tables (i.e., ``GenFactor`` assignments) for each generic factor.

        Returns
        -------
        None
        """
        if factor_assignment is None:
            factor_assignment = dict()
        self._inner.sanity_check(pub_assignment, var_assignment, factor_assignment)

    def vars(self) -> Sequence[str]:
        """Return the names of the variables in the graph."""
        return self._inner.var_names()

    def factors(self) -> Sequence[str]:
        """Return the names of the factors in the graph."""
        return self._inner.factor_names()


class BPState:
    """Belief propagation state associated to a :class:`FactorGraph`.

    This is a stateful object on which belief propagation operations can be run.
    See :class:`scalib.attacks.FactorGraph` for usage example.
    """

    def __init__(
        self,
        factor_graph: FactorGraph,
        nexec: int,
        public_values: Optional[ValsAssign] = None,
        gen_factors: Optional[GenFactors] = None,
    ):
        if public_values is None:
            public_values = dict()
        if gen_factors is None:
            gen_factors = dict()
        self._fg = factor_graph
        self._inner = factor_graph._inner.new_bp(nexec, public_values, gen_factors)

    @property
    def fg(self) -> FactorGraph:
        """The associated factor graph."""
        return self._fg

    def set_evidence(self, var: str, distribution: Optional[npt.NDArray[np.float64]]):
        r"""Sets prior distribution of a variable.

        Parameters
        ----------
        var :
            Identifier of the variable to assign the distribution to.
        distribution :
            Distribution to assign. If `var` is SINGLE, must be of shape `(nc,)`.
            If `var` is MULTI, must be of shape `(nexec,nc)`.
            If None, sets the prior distribution to uniform.
        """
        if distribution is None:
            self._inner.drop_evidence(var)
        else:
            self._inner.set_evidence(var, distribution)

    def bp_loopy(
        self,
        it: int,
        initialize_states: bool,
        clear_beliefs: bool = True,
    ):
        """Runs belief propagation algorithm on the current state of the graph.

        This is a shortcut for calls to :meth:`propagate_var` and :meth:`propagate_factor`. It is equivalent to:

        .. code-block:: python

            if initialize_states:
                for var in self.fg.vars():
                    self.propagate_var(var)
            for _ in range(it):
                for factor in self.fg.factors():
                    self.propagate_factor(factor)
                for var in self.fg.vars():
                    self.propagate_var(var)

        Parameters
        ----------
        it :
            Number of iterations of belief propagation.
        initialize_states:
            Whether to update variable distributions before running the BP iterations.
            Recommended after using :func:`BPState.set_evidence`.
        clear_beliefs:
            Whether to clear beliefs between vars -> factors. Setting to False can help debugging. Default value is True.
        """
        if initialize_states:
            self._inner.propagate_all_vars(get_config(), clear_beliefs)
        self._inner.propagate_loopy_step(it, get_config(), clear_beliefs)

    def bp_acyclic(
        self,
        dest: str,
        *,
        clear_intermediates: bool = True,
        clear_evidence: bool = False,
    ):
        """Runs the non-loopy belief propagation algorithm on the current state of the graph.
        This only works if the graph is not cyclic.

        Parameters
        ----------
        dest:
            Variable for which the belief propagation is computed.
        clear_intermediates:
            Drop the intermetidate distributions and beliefs that are computed.
        clear_evidence:
            Drop the evidence for the variables, once used in the algorithm.
        """
        self._inner.propagate_acyclic(
            dest, clear_intermediates, clear_evidence, get_config()
        )

    def get_distribution(self, var: str) -> Optional[npt.NDArray[np.float64]]:
        r"""Returns the current distribution of a variable `var`.

        Parameters
        ----------
        var : string
            Identifier of the variable for which distribution must be returned.
            Distribution cannot be obtained for public variables.

        Returns
        -------
        distribution : array_like, f64
            Distribution of `var`. If `var` is SINGLE, distribution has shape
            `(nc)`. Else, it has shape `(n,nc)`.
            If the variable has a uniform distribution, None may be returned
            (but this is not guaranteed).
        """
        return self._inner.get_state(var)

    def is_cyclic(self) -> bool:
        """Test is the graph is cyclic."""
        return self._inner.is_cyclic()

    def set_distribution(
        self, var: str, distribution: Optional[npt.NDArray[np.float64]]
    ):
        r"""Sets current distribution of a variable in the BP.

        Parameters
        ----------
        var :
            Identifier of the variable to assign the distribution to.
        distribution :
            Distribution to assign. If `var` is SINGLE, must be of shape `(nc,)`.
            If `var` is MULTI, must be of shape `(nexec,nc)`.
            If None, sets the distribution to uniform.
        """
        if distribution is None:
            self._inner.drop_state(var)
        else:
            self._inner.set_state(var, distribution)

    def get_belief_to_var(
        self, var: str, factor: str
    ) -> Optional[npt.NDArray[np.float64]]:
        r"""Returns the current belief from factor to var.

        Parameters
        ----------
        var : string
            Identifier of the variable for which distribution must be returned.
        factor : string
            Identifier of the factor for which distribution must be returned.

        Returns
        -------
        distribution : array_like, f64
            Belief on the edge from `factor` to `var`. If `factor` is SINGLE, distribution has shape
            `(nc)`. Else, it has shape `(n,nc)`.
            If the belief is a uniform distribution, None may be returned
            (but this is not guaranteed).
        """
        return self._inner.get_belief_to_var(var, factor)

    def get_belief_from_var(
        self, var: str, factor: str
    ) -> Optional[npt.NDArray[np.float64]]:
        r"""Returns the current belief from var to factor.

        Parameters
        ----------
        var : string
            Identifier of the variable for which distribution must be returned.
        factor : string
            Identifier of the factor for which distribution must be returned.

        Returns
        -------
        distribution : array_like, f64
            Belief on the edge from `var` to `factor`. If `factor` is SINGLE, distribution has shape
            `(nc)`. Else, it has shape `(n,nc)`.
            If the belief is a uniform distribution, None may be returned
            (but this is not guaranteed).
        """
        return self._inner.get_belief_from_var(var, factor)

    def propagate_var(self, var: str, clear_beliefs: bool = True):
        """Run belief propagation on variable var.

        This fetches beliefs from adjacent factors, computes the var
        distribution, and sends updated beliefs to all adjacent factors.

        Parameters
        ----------
        var : string
            Identifier of the variable.
        clear_beliefs:
            Whether to clear beliefs between vars -> factors. Setting to False can help debugging. Default value is True.

        """
        return self._inner.propagate_var(var, get_config(), clear_beliefs)

    def propagate_factor(self, factor: str):
        """Run belief propagation on the given factor.

        This fetches beliefs from adjacent variables and sends updated beliefs
        to all adjacent variables.

        Parameters
        ----------
        factor:
            Identifier of the variable.

        """
        return self._inner.propagate_factor_all(factor, get_config())

    def debug(self):
        """Debug-print the current state."""
        s = []
        s.append("VAR DISTRIBUTION")
        for var in self._inner.graph().var_names():
            s.append(f"\tVar {var}")
            s.append(repr(self.get_distribution(var)))
        s.append("VARS TO FACTORS")
        for factor in self._inner.graph().factor_names():
            for var in self._inner.graph().factor_scope(factor):
                s.append(f"\t{var} -> {factor}")
                s.append(repr(self.get_belief_from_var(var, factor)))
        s.append("FACTORS TO VARS")
        for factor in self._inner.graph().factor_names():
            for var in self._inner.graph().factor_scope(factor):
                s.append(f"\t{factor} -> {var}")
                s.append(repr(self.get_belief_to_var(var, factor)))
        return "\n".join(s)



import pytest
from scalib.attacks import FactorGraph, BPState, GenFactor
import numpy as np
import os
import copy
import itertools as it


def normalize_distr(x):
    return x / x.sum(axis=-1, keepdims=True)


def make_distri(nc, n):
    return normalize_distr(np.random.randint(1, 10000000, (n, nc)).astype(np.float64))


def test_copy_fg():
    graph = """
        NC 2
        PROPERTY s1: x = a^b
        VAR MULTI x
        VAR MULTI a
        VAR MULTI b
        """
    graph = FactorGraph(graph)
    graph2 = copy.deepcopy(graph)


def test_copy_bp():
    graph = """
        NC 2
        PROPERTY s1: x = a^b
        VAR MULTI x
        VAR MULTI a
        VAR MULTI b
        """
    graph = FactorGraph(graph)
    n = 5
    bp_state = BPState(graph, n)
    distri_a = make_distri(2, 5)
    distri_b = make_distri(2, 5)
    bp_state.set_evidence("a", distri_a)
    bp_state.set_evidence("b", distri_b)
    bp_state.bp_loopy(1, initialize_states=True)
    bp_state2 = copy.deepcopy(bp_state)


def test_table():
    """
    Test Table lookup
    """
    nc = 16
    n = 100
    table = np.random.permutation(nc).astype(np.uint32)
    distri_x = np.random.randint(1, 2048, (n, nc))
    distri_x = (distri_x.T / np.sum(distri_x, axis=1)).T

    graph = f"""
            PROPERTY y = table[x]
            TABLE table
            VAR MULTI x
            VAR MULTI y
            NC {nc}
            """
    graph = FactorGraph(graph, {"table": table})
    x = distri_x.argmax(axis=1)
    graph.sanity_check({}, {"x": x, "y": table[x]})
    bp_state = BPState(graph, n)

    bp_state.set_evidence("x", distri_x)

    bp_state.bp_loopy(1, True)
    distri_y = bp_state.get_distribution("y")

    distri_y_ref = np.zeros(distri_x.shape)
    for x in range(nc):
        y = table[x]
        distri_y_ref[:, y] = distri_x[:, x]

    assert np.allclose(distri_y_ref, distri_y)


def test_table_non_bij():
    """
    Test non-bijective Table lookup
    """
    nc = 2
    n = 1
    table = np.array([0, 0], dtype=np.uint32)
    distri_x = np.array([[0.5, 0.5]])
    distri_y = np.array([[0.8, 0.2]])

    graph = f"""
            PROPERTY y = table[x]
            TABLE table
            VAR MULTI x
            VAR MULTI y
            NC {nc}
            """
    graph = FactorGraph(graph, {"table": table})

    graph.sanity_check({}, {"x": np.array([0, 1]), "y": np.array([0, 0])})

    bp_state = BPState(graph, n)
    bp_state.set_evidence("x", distri_x)
    bp_state.set_evidence("y", distri_y)

    bp_state.bp_loopy(1, True)
    distri_x = bp_state.get_distribution("x")
    distri_y = bp_state.get_distribution("y")

    distri_x_ref = np.array([0.5, 0.5])
    distri_y_ref = np.array([1.0, 0.0])

    assert np.allclose(distri_x_ref, distri_x)
    assert np.allclose(distri_y_ref, distri_y)


def test_not():
    """
    Test NOT operation
    """
    nc = 2
    n = 1
    distri_x = np.array([[0.6, 0.4]])
    distri_y = np.array([[0.8, 0.2]])

    graph = f"""
            PROPERTY y = !x
            VAR MULTI x
            VAR MULTI y
            NC {nc}
            """
    graph = FactorGraph(graph)

    graph.sanity_check({}, {"x": np.array([0, 1]), "y": np.array([1, 0])})

    bp_state = BPState(graph, n)

    bp_state.set_evidence("x", distri_x)
    bp_state.set_evidence("y", distri_y)

    bp_state.bp_loopy(1, True)
    distri_x_bp = bp_state.get_distribution("x")
    distri_y_bp = bp_state.get_distribution("y")

    t = np.array([1, 0])
    distri_x_ref = normalize_distr(distri_x * distri_y[0, t][np.newaxis, :])
    distri_y_ref = normalize_distr(distri_y * distri_x[0, t][np.newaxis, :])

    assert np.allclose(distri_x_ref, distri_x_bp)
    assert np.allclose(distri_y_ref, distri_y_bp)


def test_and_public():
    """
    Test AND with public data
    """
    nc = 16
    n = 100
    public = np.random.randint(0, nc, n, dtype=np.uint32)
    distri_x = np.random.randint(1, 100, (n, nc))
    distri_x = (distri_x.T / np.sum(distri_x, axis=1)).T

    graph = f"""
        # some comments
        NC {nc}
        PROPERTY y = x & p
        VAR MULTI y
        VAR MULTI x
        PUB MULTI p#come comments
        """
    graph = FactorGraph(graph)
    bp_state = BPState(graph, n, {"p": public})
    bp_state.set_evidence("x", distri_x)

    bp_state.bp_loopy(2, False)

    distri_y = bp_state.get_distribution("y")
    distri_y_ref = np.zeros(distri_x.shape)
    for x in range(nc):
        y = x & public
        distri_y_ref[np.arange(n), y] += distri_x[np.arange(n), x]

    assert np.allclose(distri_y_ref, distri_y)

    bp_state2 = BPState(graph, n, {"p": public})
    bp_state2.set_evidence("x", distri_x)
    bp_state2.bp_acyclic("y")
    assert np.allclose(distri_y_ref, bp_state2.get_distribution("y"))


def test_xor_public():
    """
    Test XOR with public data
    """
    nc = 16
    n = 100
    public = np.random.randint(0, nc, n, dtype=np.uint32)
    public2 = np.random.randint(0, nc, n, dtype=np.uint32)
    distri_x = np.random.randint(1, 100, (n, nc))
    distri_x = (distri_x.T / np.sum(distri_x, axis=1)).T

    graph = f"""
        PROPERTY y = x ^ p ^ p2
        VAR MULTI y
        VAR MULTI x
        PUB MULTI p
        PUB MULTI p2
        NC {nc}
        """
    graph = FactorGraph(graph)
    bp_state = BPState(graph, n, {"p": public, "p2": public2})
    bp_state.set_evidence("x", distri_x)

    bp_state.bp_loopy(2, False)

    distri_y = bp_state.get_distribution("y")
    distri_y_ref = np.zeros(distri_x.shape)
    for x in range(nc):
        y = x ^ public ^ public2
        distri_y_ref[np.arange(n), y] = distri_x[np.arange(n), x]

    distri_x = bp_state.get_distribution("x")

    print("p ^ p2")
    print(public ^ public2)
    print("y_ref")
    print(distri_y_ref)
    print("y")
    print(distri_y)
    print("x")
    print(distri_x)
    assert distri_x is not None
    assert distri_y is not None
    assert np.allclose(distri_y_ref, distri_y)


def test_AND():
    """
    Test AND between distributions
    """
    cases = [
        (
            np.array([[0.5, 0.5]]),  # uniform x
            np.array([[0.0, 1.0]]),  # y == 1
            np.array([[0.0, 1.0]]),  # z == 1
        ),
        (
            np.array([[0.1, 0.9]]),  # x == 1
            np.array([[0.5, 0.5]]),  # uniform y
            np.array([[1.0, 0.0]]),  # z == 1
        ),
        (
            np.array([[0.1, 0.9]]),  # x == 1
            np.array([[0.4, 0.6]]),  # uniform y
            np.array([[0.8, 0.2]]),  # z == 1
        ),
        (make_distri(2, 1) for _ in range(3)),
        (make_distri(4, 1) for _ in range(3)),
        (make_distri(256, 4) for _ in range(3)),
    ]

    for distri_x, distri_y, distri_z in cases:
        print("#### Test case:")
        print(distri_x)
        print(distri_y)
        print(distri_z)
        n, nc = distri_x.shape
        graph = f"""
            # some comments
            NC {nc}
            PROPERTY z = x&y
            VAR MULTI z
            VAR MULTI x
            VAR MULTI y

            """
        graph = FactorGraph(graph)
        bp_state = BPState(graph, n)
        bp_state.set_evidence("x", distri_x)
        bp_state.set_evidence("y", distri_y)
        bp_state.set_evidence("z", distri_z)

        distri_x_ref = np.zeros(distri_x.shape)
        distri_y_ref = np.zeros(distri_y.shape)
        distri_z_ref = np.zeros(distri_z.shape)

        for x in range(nc):
            for y in range(nc):
                distri_x_ref[:, x] += distri_z[:, x & y] * distri_y[:, y]
                distri_y_ref[:, y] += distri_z[:, x & y] * distri_x[:, x]
                distri_z_ref[:, x & y] += distri_x[:, x] * distri_y[:, y]

        distri_x_ref = normalize_distr(distri_x_ref * distri_x)
        distri_y_ref = normalize_distr(distri_y_ref * distri_y)
        distri_z_ref = normalize_distr(distri_z_ref * distri_z)

        print("#### Ref:")
        print(distri_x_ref)
        print(distri_y_ref)
        print(distri_z_ref)

        bp_state.bp_loopy(1, True)
        distri_x = bp_state.get_distribution("x")
        distri_y = bp_state.get_distribution("y")
        distri_z = bp_state.get_distribution("z")

        print("#### Got:")
        print(distri_x)
        print(distri_y)
        print(distri_z)

        assert np.allclose(distri_z_ref, distri_z)
        assert np.allclose(distri_x_ref, distri_x)
        assert np.allclose(distri_y_ref, distri_y)


def test_and_not_or():
    nc = 16
    n = 100
    nc = 2
    n = 1
    p = np.random.randint(0, nc, n, dtype=np.uint32)
    t = np.random.randint(0, nc, n, dtype=np.uint32)
    nt = (nc - 1) ^ t
    distri_x = make_distri(nc, n)
    distri_y = make_distri(nc, n)
    distri_z = make_distri(nc, n)

    graph = f"""
        NC {nc}
        VAR MULTI x
        VAR MULTI y
        VAR MULTI z
        PUB MULTI p
        PUB MULTI t
        PROPERTY z = !y & x & p & !t
        """
    graph2 = f"""
        NC {nc}
        VAR MULTI x
        VAR MULTI y
        VAR MULTI ny
        PROPERTY ny = !y
        VAR MULTI z
        VAR MULTI nz
        PROPERTY z = !nz
        PUB MULTI p
        PUB MULTI nt
        VAR MULTI a
        VAR MULTI na
        PROPERTY na = !a
        VAR MULTI b
        VAR MULTI nb
        PROPERTY nb = !b
        PROPERTY a = x & p
        PROPERTY b = nt & ny 
        PROPERTY nz = na | nb 
        """
    graph = FactorGraph(graph)
    bp_state = BPState(graph, n, {"p": p, "t": t})
    bp_state.set_evidence("x", distri_x)
    bp_state.set_evidence("y", distri_y)
    bp_state.set_evidence("z", distri_z)
    bp_state.bp_loopy(1, True)
    import sys

    print("BP2", file=sys.stderr)

    graph2 = FactorGraph(graph2)
    bp_state2 = BPState(graph2, n, {"p": p, "nt": nt})
    bp_state2.set_evidence("x", distri_x)
    bp_state2.set_evidence("y", distri_y)
    bp_state2.set_evidence("z", distri_z)
    bp_state2.bp_loopy(8, False)

    for v in ["x", "y", "z"]:
        print(v)
        d = bp_state.get_distribution(v)
        d2 = bp_state2.get_distribution(v)
        assert np.allclose(d, d2)
    bp_state3 = BPState(graph, n, {"p": p, "t": t})
    bp_state3.set_evidence("x", distri_x)
    bp_state3.set_evidence("y", distri_y)
    bp_state3.set_evidence("z", distri_z)
    for v in ["x", "y", "z"]:
        print(v)
        bp_state3.bp_acyclic(v)
        assert np.allclose(bp_state2.get_distribution(v), bp_state3.get_distribution(v))


def test_ADD():
    """
    Test ADD between distributions
    """
    nc = 241
    n = 4
    distri_x = make_distri(nc, n)
    distri_y = make_distri(nc, n)

    graph = f"""
        # some comments
        NC {nc}
        PROPERTY F0: z = x+y
        VAR MULTI z
        VAR MULTI x
        VAR MULTI y

        """

    graph = FactorGraph(graph)
    bp_state = BPState(graph, n)
    bp_state.set_evidence("x", distri_x)
    bp_state.set_evidence("y", distri_y)

    bp_state.bp_loopy(1, True)
    distri_z = bp_state.get_distribution("z")

    distri_z_ref = np.zeros(distri_z.shape)
    msg = np.zeros(distri_z.shape)

    for x in range(nc):
        for y in range(nc):
            distri_z_ref[:, (x + y) % nc] += distri_x[:, x] * distri_y[:, y]

    distri_z_ref = (distri_z_ref.T / np.sum(distri_z_ref, axis=1)).T
    assert np.allclose(distri_z_ref, distri_z)


def test_ADD2():
    """
    Test ADD between distributions
    """
    nc = 16
    n = 1
    distri_x = np.zeros((n, nc))  # make_distri(nc, n)
    distri_y = np.zeros((n, nc))  # make_distri(nc, n)

    distri_x[:, 0] = 1.0
    distri_y[:, 0] = 1.0

    graph = f"""
        # some comments
        NC {nc}
        PROPERTY F: z = x+y
        VAR MULTI z
        VAR MULTI x
        VAR MULTI y

        """

    graph = FactorGraph(graph)
    bp_state = BPState(graph, n)
    bp_state.set_evidence("x", distri_x)
    bp_state.set_evidence("y", distri_y)
    print(bp_state.debug())
    bp_state.bp_loopy(1, True)
    print(bp_state.debug())
    bp_state.bp_loopy(10, False)
    print(bp_state.debug())
    distri_z = bp_state.get_distribution("z")

    distri_z_ref = np.zeros(distri_z.shape)
    msg = np.zeros(distri_z.shape)

    for x in range(nc):
        for y in range(nc):
            distri_z_ref[:, (x + y) % nc] += distri_x[:, x] * distri_y[:, y]

    distri_z_ref = (distri_z_ref.T / np.sum(distri_z_ref, axis=1)).T
    assert np.allclose(distri_z_ref, distri_z)


def test_add_cst():
    """
    Test ADD of a constant
    """
    nc = 251
    y_cases = [0, 1, 2, 58, 249, 250, 251, 2345]
    n = len(y_cases)
    distri_x = make_distri(nc, n)

    graph = f"""
        # some comments
        NC {nc}
        PROPERTY z = x+y+w
        VAR MULTI z
        VAR MULTI x
        PUB MULTI y
        PUB SINGLE w
        """
    y_cases = np.array(y_cases, dtype=np.uint32)
    graph = FactorGraph(graph)
    w = 5
    for evidence_var, target_var, sub in [("x", "z", False), ("z", "x", True)]:
        print("sub", sub)
        bp_state = BPState(graph, n, {"y": y_cases, "w": w})
        bp_state.set_evidence(evidence_var, distri_x)

        bp_state.bp_loopy(1, True)
        distri_z = bp_state.get_distribution(target_var)

        distri_z_ref = np.zeros(distri_z.shape)

        for x in range(nc):
            for j in range(n):
                if sub:
                    new_idx = (x - y_cases[j] - w) % nc
                else:
                    new_idx = (x + y_cases[j] + w) % nc
                distri_z_ref[j, new_idx] = distri_x[j, x]

        assert np.allclose(distri_z_ref, distri_z)


def test_ADD_multiple():
    """
    Test ADD between distributions
    """
    nc = 17
    n = 4
    distri_x = make_distri(nc, n)
    distri_y = make_distri(nc, n)
    distri_w = make_distri(nc, n)

    graph = f"""
        # some comments
        NC {nc}
        PROPERTY z = x+y+w
        VAR MULTI z
        VAR MULTI x
        VAR MULTI y
        VAR MULTI w
        """

    graph = FactorGraph(graph)
    bp_state = BPState(graph, n)
    bp_state.set_evidence("x", distri_x)
    bp_state.set_evidence("y", distri_y)
    bp_state.set_evidence("w", distri_w)

    bp_state.bp_loopy(1, True)
    distri_z = bp_state.get_distribution("z")

    distri_z_ref = np.zeros(distri_z.shape)
    msg = np.zeros(distri_z.shape)

    for x in range(nc):
        for y in range(nc):
            for w in range(nc):
                distri_z_ref[:, (x + y + w) % nc] += (
                    distri_x[:, x] * distri_y[:, y] * distri_w[:, w]
                )

    distri_z_ref = (distri_z_ref.T / np.sum(distri_z_ref, axis=1)).T
    assert np.allclose(distri_z_ref, distri_z)


def test_add_cst2():
    """More test cases for addition with public values"""
    nc = 256
    n = 1
    graph = f"""
        NC {nc}
        PROPERTY z = -x+y+w
        PUB MULTI z
        VAR MULTI x
        VAR MULTI y
        PUB SINGLE w
        """
    graph = FactorGraph(graph)
    bp_state = BPState(graph, n, {"z": [2], "w": 1})
    distri_x = np.zeros((1, nc))
    distri_x[0, 4] = 1.0
    bp_state.set_evidence("x", distri_x)
    distri_y_ref = np.zeros((1, nc))
    # y = z + x - w = 2 + 4 - 1 = 5
    distri_y_ref[0, 5] = 1.0
    bp_state.bp_acyclic("y")
    distri_y = bp_state.get_distribution("y")
    assert np.allclose(distri_y_ref, distri_y)


def test_add_cst3():
    """More even more cases for addition with public values"""
    nc = 256
    g_pub = f"""
    NC {nc}
    PUB SINGLE s
    VAR MULTI a0
    VAR MULTI a1
    PROPERTY s = a0 + a1
    """
    g_var = f"""
    NC {nc}
    VAR SINGLE s
    VAR MULTI a0
    VAR MULTI a1
    PROPERTY s = a0 + a1
    """
    n = 1
    fg_pub = FactorGraph(g_pub)
    fg_var = FactorGraph(g_var)
    one_distr = np.array([[1.0 if i == 1 else 0.0 for i in range(nc)]])
    bp_pub = BPState(fg_pub, n, {"s": 1})
    bp_pub.set_evidence("a0", one_distr)
    bp_var = BPState(fg_var, n)
    bp_var.set_evidence("s", one_distr[0])
    bp_var.set_evidence("a0", one_distr)
    bp_pub.bp_acyclic("a1")
    bp_var.bp_acyclic("a1")
    pub_res = bp_pub.get_distribution("a1")
    var_res = bp_var.get_distribution("a1")
    assert np.allclose(pub_res, var_res)


def test_MUL():
    """
    Test MUL between distributions
    """
    nc = 251
    n = 4
    distri_x = make_distri(nc, n)
    distri_y = make_distri(nc, n)

    graph = f"""
        # some comments
        NC {nc}
        PROPERTY z = x*y
        VAR MULTI z
        VAR MULTI x
        VAR MULTI y

        """

    graph = FactorGraph(graph)
    bp_state = BPState(graph, n)
    bp_state.set_evidence("x", distri_x)
    bp_state.set_evidence("y", distri_y)

    bp_state.bp_loopy(1, True)
    distri_z = bp_state.get_distribution("z")

    distri_z_ref = np.zeros(distri_z.shape)
    msg = np.zeros(distri_z.shape)

    for x in range(nc):
        for y in range(nc):
            distri_z_ref[:, (x * y) % nc] += distri_x[:, x] * distri_y[:, y]

    distri_z_ref = (distri_z_ref.T / np.sum(distri_z_ref, axis=1)).T
    assert np.allclose(distri_z_ref, distri_z)


def test_xor():
    """
    Test XOR between distributions
    """
    nc = 512
    n = 10
    distri_x = np.random.randint(1, 10000000, (n, nc))
    distri_x = (distri_x.T / np.sum(distri_x, axis=1)).T
    distri_y = np.random.randint(1, 10000000, (n, nc))
    distri_y = (distri_y.T / np.sum(distri_y, axis=1)).T

    distri_b = np.random.randint(1, 10000000, (n, nc))
    distri_b[1, 0] = 1.0
    distri_b[1, 1] = 0.0
    distri_b = (distri_b.T / np.sum(distri_b, axis=1)).T
    distri_a = np.random.randint(1, 10000000, (n, nc))
    distri_a[1, :] = 1.0
    distri_a = (distri_a.T / np.sum(distri_a, axis=1)).T

    graph = f"""
        # some comments
        NC {nc}
        PROPERTY s2: z = x^y
        PROPERTY s1: x = a^b
        VAR MULTI x # some comments too
        VAR MULTI y
        VAR MULTI a
        VAR MULTI b
        VAR SINGLE z"""

    graph = FactorGraph(graph)
    bp_state = BPState(graph, n)
    bp_state.set_evidence("x", distri_x)
    bp_state.set_evidence("y", distri_y)
    bp_state.set_evidence("a", distri_a)
    bp_state.set_evidence("b", distri_b)

    bp_state2 = copy.deepcopy(bp_state)

    # Custom optimized sequence -- equivelent to non-loopy algo, but focusing only on z.
    bp_state.propagate_var("a")
    bp_state.propagate_var("b")
    bp_state.propagate_var("x")
    bp_state.propagate_factor("s1")
    bp_state.propagate_var("x")
    bp_state.propagate_var("y")
    bp_state.propagate_factor("s2")
    bp_state.propagate_var("z")

    bp_state2.bp_loopy(2, True)

    distri_z = bp_state.get_distribution("z")
    distri_z2 = bp_state2.get_distribution("z")

    msg = np.zeros((n, nc))
    distri_z_ref_multi = np.zeros((n, nc))
    distri_z_ref = np.ones(distri_z.shape)

    for a in range(nc):
        for b in range(nc):
            msg[:, a ^ b] += distri_a[:, a] * distri_b[:, b]

    distri_x *= msg
    for x in range(nc):
        for y in range(nc):
            distri_z_ref_multi[:, x ^ y] += distri_x[:, x] * distri_y[:, y]
    for d in distri_z_ref_multi:
        print(d)
        distri_z_ref *= d

    print("distri_x_ref", distri_x)
    print("distri_y_ref", distri_y)

    distri_z_ref = distri_z_ref / np.sum(distri_z_ref)
    print("distri_z_ref", distri_z_ref)
    assert np.allclose(distri_z_ref, distri_z)
    assert np.allclose(distri_z_ref, distri_z2)


def test_xor_acyclic():
    nc = 4
    fg = FactorGraph(
        f"""NC {nc}
    VAR MULTI x0
    VAR MULTI x4
    VAR MULTI y0
    PROPERTY F1: y0 = x0 ^ x4
    """
    )
    bp = BPState(fg, 1)
    for v in fg.vars():
        bp.set_evidence(v, make_distri(nc, 1))
    for v in ["x0", "x4"]:
        bp.bp_acyclic(v)
        bp.get_distribution(v)


def test_bad_var_norm():
    """
    When normalization of distributions can be bad...
    Root cause of bug #55.
    """
    nc = 2
    n = 1
    dy = np.array([[1.0, 1e-100]])
    dz = np.array([[1e-100, 1.0]])
    graph = f"""
        NC {nc}
        VAR MULTI x
        VAR MULTI y
        VAR MULTI z
        PUB MULTI a
        PROPERTY s1: x = y^a
        PROPERTY s2: x = !z
        """

    graph = FactorGraph(graph)
    bp_state = BPState(graph, n, {"a": [0]})
    bp_state.set_evidence("y", dy)
    bp_state.set_evidence("z", dz)

    bp_state.propagate_var("y")
    bp_state.propagate_var("z")
    bp_state.propagate_factor("s1")
    bp_state.propagate_factor("s2")
    bp_state.propagate_var("x")

    s1 = normalize_distr(bp_state.get_belief_from_var("x", "s1"))
    s2 = normalize_distr(bp_state.get_belief_from_var("x", "s2"))

    assert np.allclose(np.array([[1.0, 1e-40]]), s1)
    assert np.allclose(np.array([[1.0, 1e-40]]), s2)


def test_cyclic():
    graph_never_cyclic = """
    NC 2
    VAR MULTI x
    VAR MULTI x0
    VAR MULTI x1
    VAR MULTI y
    VAR MULTI y0
    VAR MULTI y1
    PUB MULTI p
    VAR SINGLE k
    VAR SINGLE k2

    PROPERTY x = p ^ k
    PROPERTY y = y0 ^ y1
    PROPERTY x = x0 ^ x1
    PROPERTY k2  = p ^ y
    """
    g = FactorGraph(graph_never_cyclic)
    assert BPState(g, 1, {"p": np.array([0], dtype=np.uint32)}).is_cyclic() == False
    assert BPState(g, 2, {"p": np.array([0, 0], dtype=np.uint32)}).is_cyclic() == False
    graph_always_cyclic = """
    NC 2
    VAR MULTI x
    VAR MULTI x0
    VAR MULTI x1
    VAR MULTI y
    VAR MULTI y0
    VAR MULTI y1
    PUB MULTI p
    VAR SINGLE k

    PROPERTY x = p ^ k
    PROPERTY y = y0 ^ y1
    PROPERTY x = x0 ^ x1
    PROPERTY y  = p ^ y1
    """
    g = FactorGraph(graph_always_cyclic)
    assert BPState(g, 1, {"p": np.array([0], dtype=np.uint32)}).is_cyclic() == True
    assert BPState(g, 2, {"p": np.array([0, 0], dtype=np.uint32)}).is_cyclic() == True
    graph_multi_cyclic = """
    NC 2
    VAR MULTI x
    VAR MULTI x0
    VAR MULTI x1
    VAR MULTI y
    VAR MULTI y0
    VAR MULTI y1
    PUB MULTI p
    VAR SINGLE k
    VAR SINGLE k2

    PROPERTY x = p ^ k
    PROPERTY y = y0 ^ y1
    PROPERTY x = x0 ^ x1
    PROPERTY k2  = x ^ y
    """
    g = FactorGraph(graph_multi_cyclic)
    assert BPState(g, 1, {"p": np.array([0], dtype=np.uint32)}).is_cyclic() == False
    assert BPState(g, 2, {"p": np.array([0, 0], dtype=np.uint32)}).is_cyclic() == True


def test_and_rounding_error_simple():
    # simple reproduction of issue #86
    factor_graph = """NC 16
    VAR MULTI A
    VAR MULTI B
    VAR MULTI C
    PROPERTY P: C = A & B
    """
    priors = {
        "A": [
            1.0 + 2**-52,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            1.0,
        ],
        "C": [
            0.0,
            2.666666666666667,
            2.666666666666667,
            0.0,
            2.666666666666667,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    }
    fg = FactorGraph(factor_graph)
    bp = BPState(fg, 1)
    for k, v in priors.items():
        bp._inner.set_belief_from_var(k, "P", np.array([v]))
    bp.propagate_factor("P")
    assert (bp.get_belief_to_var("B", "P") >= 0.0).all()


def test_and_rounding_error_simple2():
    # test case of issue #86
    factor_graph = """NC 16
    VAR MULTI K0
    VAR MULTI K1
    VAR MULTI L1
    VAR MULTI N1
    VAR MULTI B
    VAR MULTI C
    VAR MULTI N0
    VAR MULTI A
    VAR MULTI L2
    VAR MULTI L3
    PUB SINGLE IV
    PROPERTY P1: L1 = K0 ^ K1
    PROPERTY P2: N1 = !K1
    PROPERTY P3: B = N1 & IV
    PROPERTY P4: C = B ^ K0
    PROPERTY P5: N0 = !K0
    PROPERTY P6: A = N0 & K1
    PROPERTY P7: L2 = A ^ C
    PROPERTY P8: L3 = K1 ^ C"""
    priors = {
        "A": [
            0.0,
            0.25,
            0.25,
            0.0,
            0.25,
            0.0,
            0.0,
            0.0,
            0.25,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        "C": [
            0.0,
            0.0,
            0.0,
            0.1666666667,
            0.0,
            0.1666666667,
            0.1666666667,
            0.0,
            0.0,
            0.1666666667,
            0.1666666667,
            0.0,
            0.1666666667,
            0.0,
            0.0,
            0.0,
        ],
        "L1": [
            0.0,
            0.0,
            0.0,
            0.1666666667,
            0.0,
            0.1666666667,
            0.1666666667,
            0.0,
            0.0,
            0.1666666667,
            0.1666666667,
            0.0,
            0.1666666667,
            0.0,
            0.0,
            0.0,
        ],
        "L2": [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.25,
            0.0,
            0.0,
            0.0,
            0.25,
            0.0,
            0.25,
            0.25,
            0.0,
        ],
        "L3": [
            0.0,
            0.0,
            0.0,
            0.1666666667,
            0.0,
            0.1666666667,
            0.1666666667,
            0.0,
            0.0,
            0.1666666667,
            0.1666666667,
            0.0,
            0.1666666667,
            0.0,
            0.0,
            0.0,
        ],
    }
    fg = FactorGraph(factor_graph)
    bp = BPState(fg, 1, public_values={"IV": 0xC})
    for k, v in priors.items():
        bp.set_evidence(k, distribution=np.array([v]))
    bp.bp_loopy(5, initialize_states=False)
    assert (bp.get_distribution("K0") >= 0.0).all()
    assert (bp.get_distribution("K1") >= 0.0).all()


def test_manytraces():
    """No numerical underflow with many traces."""
    nc = 8
    n = 500
    distri_x = np.ones((n, nc))
    distri_x[:, 0] = 2.0
    distri_x[:, 1] = 1.5
    distri_x = (distri_x.T / np.sum(distri_x, axis=1)).T

    graph = f"""
            NC {nc}
            VAR MULTI x
            VAR SINGLE y
            PUB SINGLE c
            PROPERTY y = x ^ c
            """
    graph = FactorGraph(graph)
    bp_state = BPState(graph, n, {"c": 0})

    bp_state.set_evidence("x", distri_x)

    bp_state.bp_acyclic("y")
    distri_y = bp_state.get_distribution("y")

    distri_y_ref = np.log2(distri_x).sum(axis=0, keepdims=True)
    distri_y_ref = distri_y_ref - np.max(distri_y_ref[0, :])
    distri_y_ref = 2**distri_y_ref
    distri_y_ref = distri_y_ref / distri_y_ref.sum(axis=1, keepdims=True)

    assert np.allclose(distri_y_ref, distri_y, rtol=1e-5, atol=1e-19)


def test_sparse_factor_xor():
    from scalib.attacks.factor_graph import GenFactor

    nc = 256
    graph = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI c
    GENERIC SINGLE f
    PROPERTY F0: f(a,b,c)"""

    graph2 = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI c
    PROPERTY F0: c = a ^ b"""
    fg = FactorGraph(graph)
    fg2 = FactorGraph(graph2)
    xor = []
    n = 4
    for a in range(nc):
        for b in range(nc):
            xor.append([a, b, (a ^ b) & 0xFF])
    bp = BPState(
        fg,
        n,
        gen_factors={"f": GenFactor.sparse_functional(np.array(xor, dtype=np.uint32))},
    )
    bp2 = BPState(fg2, n)
    distr_a = make_distri(nc, n)
    distr_b = make_distri(nc, n)
    bp.set_evidence("a", distr_a)
    bp.set_evidence("b", distr_b)
    bp2.set_evidence("a", distr_a)
    bp2.set_evidence("b", distr_b)

    bp.bp_loopy(10, False)
    bp2.bp_loopy(10, False)
    assert np.allclose(bp.get_distribution("c"), bp2.get_distribution("c"))


def test_sparse_factor_xor_multi():
    from scalib.attacks.factor_graph import GenFactor

    nc = 256
    graph = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI c
    GENERIC MULTI f
    PROPERTY F0: f(a,b,c)"""

    graph2 = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI c
    PROPERTY F0: c = a ^ b"""
    fg = FactorGraph(graph)
    fg2 = FactorGraph(graph2)
    xor = []
    n = 4
    for a in range(nc):
        for b in range(nc):
            xor.append([a, b, (a ^ b) & 0xFF])
    bp = BPState(
        fg,
        n,
        gen_factors={
            "f": [
                GenFactor.sparse_functional(np.array(xor, dtype=np.uint32))
                for _ in range(n)
            ]
        },
    )
    bp2 = BPState(fg2, n)
    distr_a = make_distri(nc, n)
    distr_b = make_distri(nc, n)
    bp.set_evidence("a", distr_a)
    bp.set_evidence("b", distr_b)
    bp2.set_evidence("a", distr_a)
    bp2.set_evidence("b", distr_b)

    bp.bp_loopy(10, False)
    bp2.bp_loopy(10, False)
    assert np.allclose(bp.get_distribution("c"), bp2.get_distribution("c"))


def test_dense_factor_bff():
    from scalib.attacks.factor_graph import GenFactor

    nc = 13
    graph = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI c
    VAR MULTI d
    GENERIC SINGLE f
    PROPERTY F0: f(a,b,c,d)"""

    graph2 = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI bi
    VAR MULTI c
    VAR MULTI d
    TABLE sub
    PROPERTY F0: c = a + b
    PROPERTY F1: bi = sub[b]
    PROPERTY F2: d = a + bi"""

    fg = FactorGraph(graph)
    fg2 = FactorGraph(
        graph2, {"sub": np.array([-n % nc for n in range(nc)], dtype=np.uint32)}
    )
    bff = np.zeros((nc, nc, nc, nc))
    n = 4
    for a in range(nc):
        for b in range(nc):
            bff[a, b, (a + b) % nc, (a - b) % nc] = 1.0

    bp = BPState(fg, n, gen_factors={"f": GenFactor.dense(bff)})
    bp2 = BPState(fg2, n)

    distr_a = make_distri(nc, n)
    distr_b = make_distri(nc, n)
    bp.set_evidence("a", distr_a)
    bp.set_evidence("b", distr_b)
    bp2.set_evidence("a", distr_a)
    bp2.set_evidence("b", distr_b)

    bp.bp_loopy(10, False)
    bp2.bp_loopy(10, False)
    assert np.allclose(bp.get_distribution("c"), bp2.get_distribution("c"))


def test_dense_factor_bff():
    from scalib.attacks.factor_graph import GenFactor

    nc = 13
    graph = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI c
    VAR MULTI d
    GENERIC MULTI f
    PROPERTY F0: f(a,b,c,d)"""

    graph2 = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI bi
    VAR MULTI c
    VAR MULTI d
    TABLE sub
    PROPERTY F0: c = a + b
    PROPERTY F1: bi = sub[b]
    PROPERTY F2: d = a + bi"""

    fg = FactorGraph(graph)
    fg2 = FactorGraph(
        graph2, {"sub": np.array([-n % nc for n in range(nc)], dtype=np.uint32)}
    )
    bff = np.zeros((nc, nc, nc, nc))
    n = 4
    for a in range(nc):
        for b in range(nc):
            bff[a, b, (a + b) % nc, (a - b) % nc] = 1.0

    bp = BPState(fg, n, gen_factors={"f": [GenFactor.dense(bff) for _ in range(n)]})
    bp2 = BPState(fg2, n)

    distr_a = make_distri(nc, n)
    distr_b = make_distri(nc, n)
    bp.set_evidence("a", distr_a)
    bp.set_evidence("b", distr_b)
    bp2.set_evidence("a", distr_a)
    bp2.set_evidence("b", distr_b)

    bp.bp_loopy(10, False)
    bp2.bp_loopy(10, False)
    assert np.allclose(bp.get_distribution("c"), bp2.get_distribution("c"))


def test_manytraces():
    """No numerical underflow with many traces."""
    nc = 8
    n = 500
    distri_x = np.ones((n, nc))
    distri_x[:, 0] = 2.0
    distri_x[:, 1] = 1.5
    distri_x = (distri_x.T / np.sum(distri_x, axis=1)).T

    graph = f"""
            NC {nc}
            VAR MULTI x
            VAR SINGLE y
            PUB SINGLE c
            PROPERTY y = x ^ c
            """
    graph = FactorGraph(graph)
    bp_state = BPState(graph, n, {"c": 0})

    bp_state.set_evidence("x", distri_x)

    bp_state.bp_acyclic("y")
    distri_y = bp_state.get_distribution("y")

    distri_y_ref = np.log2(distri_x).sum(axis=0, keepdims=True)
    distri_y_ref = distri_y_ref - np.max(distri_y_ref[0, :])
    distri_y_ref = 2**distri_y_ref
    distri_y_ref = distri_y_ref / distri_y_ref.sum(axis=1, keepdims=True)

    assert np.allclose(distri_y_ref, distri_y, rtol=1e-5, atol=1e-19)


def test_ADD3():
    nc = 13
    graph = f"""NC {nc}
    TABLE SUB = [0, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1]
    VAR MULTI a0
    VAR MULTI x0
    VAR MULTI x1



    PROPERTY F0: a0 = x0 + x1
    """
    a0_distr = np.array(
        [[1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]
    )
    x0_distr = np.array(
        [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0]]
    )
    x1_distr = np.array(
        [[0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]
    )
    fg = FactorGraph(graph)
    bp_state = BPState(fg, 1)
    bp_state.set_evidence("x0", x0_distr)
    bp_state.set_evidence("x1", x1_distr)
    bp_state.set_evidence("a0", a0_distr)
    bp_state.bp_loopy(50, initialize_states=False)
    for x in ["x0", "x1", "a0"]:
        print(bp_state.get_distribution(x))
        assert not np.isnan(bp_state.get_distribution(x)).any()


def test_ADD4():
    nc = 256
    n = 10
    graph = f"""NC {nc}
    VAR MULTI a0
    VAR MULTI x0
    VAR MULTI x1



    PROPERTY F0: a0 = x0 + x1
    """
    fg = FactorGraph(graph)
    for _ in range(50):
        bp = BPState(fg, n)

        bp.set_evidence("x0", make_distri(nc, n))
        bp.set_evidence("x1", make_distri(nc, n))
        bp.set_evidence("a0", make_distri(nc, n))
        bp.bp_loopy(50, initialize_states=False)
        for x in ["x0", "x1", "a0"]:
            assert not np.isnan(bp.get_distribution(x)).any()


def test_mix_single_multi():
    graph_desc = """
    NC 2
    TABLE t
    VAR SINGLE s

    VAR MULTI a
    VAR MULTI b

    PROPERTY TABLE: a = t[s]
    PROPERTY XOR: b = a ^ s
    """

    graph_desc2 = """
    NC 2
    TABLE t
    VAR SINGLE s

    VAR SINGLE a0
    VAR SINGLE b0

    PROPERTY a0 = t[s]
    PROPERTY b0 = a0 ^ s

    VAR SINGLE a1
    VAR SINGLE b1

    PROPERTY a1 = t[s]
    PROPERTY b1 = a1 ^ s
    """

    graph = FactorGraph(graph_desc, {"t": np.arange(2, dtype=np.uint32)})
    graph2 = FactorGraph(graph_desc2, {"t": np.arange(2, dtype=np.uint32)})

    sasca = BPState(graph, 2)
    sasca.bp_loopy(4, initialize_states=True)  # check no crash

    sasca = BPState(graph, 2)
    sasca2 = BPState(graph2, 1)

    for _ in range(10):
        sasca.bp_loopy(1, initialize_states=False)
        sasca2.bp_loopy(1, initialize_states=False)
        for v in ("a", "b"):
            d = sasca.get_distribution(v)
            d0 = sasca2.get_distribution(f"{v}0")
            d1 = sasca2.get_distribution(f"{v}0")
            if d is not None:
                assert np.allclose(d[0, :], d0, rtol=1e-5, atol=1e-19)
                assert np.allclose(d[1, :], d1, rtol=1e-5, atol=1e-19)
        d = sasca.get_distribution("s")
        d2 = sasca2.get_distribution("s")
        if d is not None:
            assert np.allclose(d, d2, rtol=1e-5, atol=1e-19)


def test_mix_single_multi2():
    graph = f"""
        NC 16
        VAR SINGLE A
        VAR MULTI B
        VAR MULTI AnB
        VAR MULTI C
        VAR MULTI AnC

        PROPERTY A_and_B: AnB = A & B
        PROPERTY A_and_C: AnC = A & C
    """
    fg = FactorGraph(graph)
    bp = BPState(fg, 2)
    bp.bp_loopy(3, initialize_states=True)


def test_ADD5():
    for nc in [13, 256]:
        n = 10
        graph = f"""NC {nc}
        VAR MULTI x
        VAR MULTI y
        VAR MULTI z



        PROPERTY F0: z = x + y
        """

        fg = FactorGraph(graph)
        for _ in range(50):
            bp = BPState(fg, n)

            x_distri = make_distri(nc, n)
            z_distri = make_distri(nc, n)
            bp.set_evidence("x", x_distri)
            bp.set_evidence("z", z_distri)

            y_distri_ref = np.zeros(x_distri.shape)

            for x in range(nc):
                for z in range(nc):
                    y_distri_ref[:, (z - x) % nc] += x_distri[:, x] * z_distri[:, z]

            y_distri_ref = (y_distri_ref.T / np.sum(y_distri_ref, axis=1)).T

            bp.bp_loopy(3, initialize_states=False)

            assert np.allclose(y_distri_ref, bp.get_distribution("y"))
            for x in ["x", "y", "z"]:
                print(bp.get_distribution(x))
                assert not np.isnan(bp.get_distribution(x)).any()


def test_clear_beliefs():
    graph = """
    NC 2
    PROPERTY s1: x = a^b
    VAR MULTI x
    VAR MULTI a
    VAR MULTI b
    """
    fg = FactorGraph(graph)
    bp = BPState(fg, 1)
    for k in fg.vars():
        bp.set_evidence(k, make_distri(2, 1))
    bp.bp_loopy(1, False, clear_beliefs=False)

    assert bp.get_belief_from_var("x", "s1") is not None
    assert bp.get_belief_from_var("a", "s1") is not None
    assert bp.get_belief_from_var("b", "s1") is not None


def test_sub():
    g = """
    NC 256
    PROPERTY s1: z = x - y
    VAR MULTI x
    VAR MULTI y
    VAR MULTI z
    """

    nc = 256
    n = 10
    fg = FactorGraph(g)
    bp = BPState(fg, n)
    distri_x = make_distri(nc, n)
    distri_y = make_distri(nc, n)
    bp.set_evidence("x", distri_x)
    bp.set_evidence("y", distri_y)
    z_distri_ref = np.zeros(distri_x.shape)

    for x in range(nc):
        for y in range(nc):
            z_distri_ref[:, (x - y) % nc] += distri_x[:, x] * distri_y[:, y]

    z_distri_ref = (z_distri_ref.T / np.sum(z_distri_ref, axis=1)).T
    bp.bp_loopy(1, True)
    distri_z = bp.get_distribution("z")

    assert np.allclose(z_distri_ref, distri_z)


def test_sub_multi_ops():
    g = """
    NC 13
    PROPERTY s1: z = -x + q - w + y
    VAR MULTI x
    VAR MULTI y
    VAR MULTI z
    VAR MULTI q
    VAR MULTI w
    """

    nc = 13
    n = 10
    fg = FactorGraph(g)
    bp = BPState(fg, n)
    distri_x = make_distri(nc, n)
    distri_y = make_distri(nc, n)
    distri_q = make_distri(nc, n)
    distri_w = make_distri(nc, n)

    bp.set_evidence("x", distri_x)
    bp.set_evidence("y", distri_y)
    bp.set_evidence("w", distri_w)
    bp.set_evidence("q", distri_q)
    z_distri_ref = np.zeros(distri_x.shape)

    for x in range(nc):
        for y in range(nc):
            for w in range(nc):
                for q in range(nc):
                    z_distri_ref[:, (-x + q - w + y) % nc] += (
                        distri_x[:, x]
                        * distri_y[:, y]
                        * distri_w[:, w]
                        * distri_q[:, q]
                    )

    z_distri_ref = (z_distri_ref.T / np.sum(z_distri_ref, axis=1)).T
    bp.bp_acyclic("z")
    distri_z = bp.get_distribution("z")

    assert np.allclose(z_distri_ref, distri_z)


def test_single_gf_sanity_check():
    from scalib.attacks.factor_graph import GenFactor

    nc = 13
    graph = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI c
    VAR MULTI d
    GENERIC SINGLE f
    PROPERTY F0: f(a,b,c,d)"""

    fg = FactorGraph(graph)

    bff = np.zeros((nc, nc, nc, nc))
    n = 1
    for a in range(nc):
        for b in range(nc):
            bff[a, b, (a + b) % nc, (a - b) % nc] = 1.0

    bff_sparse = []
    for a in range(nc):
        for b in range(nc):
            bff_sparse.append([a, b, (a + b) % nc, (a - b) % nc])

    fg.sanity_check(
        {},
        {
            "a": np.array([0, 0]),
            "b": np.array([1, 1]),
            "c": np.array([1, 1]),
            "d": np.array([12, 12]),
        },
        {"f": GenFactor.dense(bff)},
    )
    fg.sanity_check(
        {},
        {
            "a": np.array([0, 0]),
            "b": np.array([1, 1]),
            "c": np.array([1, 1]),
            "d": np.array([12, 12]),
        },
        {"f": GenFactor.sparse_functional(np.array(bff_sparse, dtype=np.uint32))},
    )


def test_multi_gf_sanity_check():
    from scalib.attacks.factor_graph import GenFactor

    nc = 13
    graph = f"""NC {nc}
    VAR MULTI a
    VAR MULTI b
    VAR MULTI c
    VAR MULTI d
    GENERIC MULTI f
    PROPERTY F0: f(a,b,c,d)"""

    fg = FactorGraph(graph)

    bff = np.zeros((nc, nc, nc, nc))
    bff2 = np.zeros((nc, nc, nc, nc))
    n = 1
    for a in range(nc):
        for b in range(nc):
            bff[a, b, (a + b) % nc, (a - b) % nc] = 1.0
            bff2[a, b, (a + 2 * b) % nc, (a - 2 * b) % nc] = 1.0

    bff_sparse = []
    bff2_sparse = []
    for a in range(nc):
        for b in range(nc):
            bff_sparse.append([a, b, (a + b) % nc, (a - b) % nc])
            bff2_sparse.append([a, b, (a + 2 * b) % nc, (a - 2 * b) % nc])
    fg.sanity_check(
        {},
        {
            "a": np.array([0, 0]),
            "b": np.array([1, 1]),
            "c": np.array([1, 2]),
            "d": np.array([12, 11]),
        },
        {"f": [GenFactor.dense(bff), GenFactor.dense(bff2)]},
    )
    fg.sanity_check(
        {},
        {
            "a": np.array([0, 0]),
            "b": np.array([1, 1]),
            "c": np.array([1, 2]),
            "d": np.array([12, 11]),
        },
        {
            "f": [
                GenFactor.sparse_functional(np.array(bff_sparse, dtype=np.uint32)),
                GenFactor.sparse_functional(np.array(bff2_sparse, dtype=np.uint32)),
            ]
        },
    )


def test_mixed_gf_sanity_check():
    from scalib.attacks.factor_graph import GenFactor

    nc = 13
    graph = f"""NC {nc}
    VAR MULTI a
    VAR SINGLE b
    VAR MULTI c
    VAR MULTI d
    GENERIC MULTI f
    PROPERTY F0: f(a,b,c,d)"""

    graph2 = f"""NC {nc}
    VAR MULTI a
    VAR SINGLE b
    VAR MULTI c
    VAR MULTI d
    GENERIC SINGLE f
    PROPERTY F0: f(a,b,c,d)"""
    fg = FactorGraph(graph)
    fg2 = FactorGraph(graph2)

    bff = np.zeros((nc, nc, nc, nc))
    bff2 = np.zeros((nc, nc, nc, nc))
    n = 1
    for a in range(nc):
        for b in range(nc):
            bff[a, b, (a + b) % nc, (a - b) % nc] = 1.0
            bff2[a, b, (a + 2 * b) % nc, (a - 2 * b) % nc] = 1.0

    bff_sparse = []
    bff2_sparse = []
    for a in range(nc):
        for b in range(nc):
            bff_sparse.append([a, b, (a + b) % nc, (a - b) % nc])
            bff2_sparse.append([a, b, (a + 2 * b) % nc, (a - 2 * b) % nc])
    fg.sanity_check(
        {},
        {
            "a": np.array([0, 0]),
            "b": 1,
            "c": np.array([1, 2]),
            "d": np.array([12, 11]),
        },
        {
            "f": [GenFactor.dense(bff), GenFactor.dense(bff2)],
        },
    )
    fg.sanity_check(
        {},
        {
            "a": np.array([0, 0]),
            "b": 1,
            "c": np.array([1, 2]),
            "d": np.array([12, 11]),
        },
        {
            "f": [
                GenFactor.sparse_functional(np.array(bff_sparse, dtype=np.uint32)),
                GenFactor.sparse_functional(np.array(bff2_sparse, dtype=np.uint32)),
            ]
        },
    )

    fg2.sanity_check(
        {},
        {"a": np.array([0, 1]), "b": 1, "c": np.array([1, 2]), "d": np.array([12, 0])},
        {"f": GenFactor.dense(bff)},
    )
    fg2.sanity_check(
        {},
        {"a": np.array([0, 1]), "b": 1, "c": np.array([1, 2]), "d": np.array([12, 0])},
        {"f": GenFactor.sparse_functional(np.array(bff_sparse, dtype=np.uint32))},
    )


def test_sanity_or():
    graph = """
    NC 2
    PROPERTY s1: x = a | b
    VAR SINGLE x
    VAR SINGLE a
    VAR MULTI b
    """
    fg = FactorGraph(graph)
    fg.sanity_check({}, {"x": 0, "a": 0, "b": [0, 0]})
    fg.sanity_check({}, {"x": 1, "a": 1, "b": [0, 0]})
    fg.sanity_check({}, {"x": 1, "a": 0, "b": [1, 1]})
    fg.sanity_check({}, {"x": 1, "a": 1, "b": [1, 0]})
    graph = """
    NC 2
    PROPERTY s1: x = a & b
    VAR MULTI x
    VAR SINGLE a
    VAR MULTI b
    """
    fg = FactorGraph(graph)
    fg.sanity_check({}, {"x": [0, 0], "a": 0, "b": [0, 0]})
    fg.sanity_check({}, {"x": [0, 0], "a": 0, "b": [0, 1]})
    fg.sanity_check({}, {"x": [0, 0], "a": 0, "b": [1, 0]})
    fg.sanity_check({}, {"x": [1, 1], "a": 1, "b": [1, 1]})


def test_cycle_detection_single_factor():
    graph_desc = """
                NC 2
                VAR SINGLE x
                VAR SINGLE y
                PROPERTY x = !y
                """

    fg = FactorGraph(graph_desc)
    bp = BPState(fg, 2)
    assert not bp.is_cyclic()


def test_cycle_detection_single_factor2():
    graph_desc = """
                NC 2
                VAR SINGLE x
                VAR SINGLE y
                VAR SINGLE z
                PROPERTY x = !y
                PROPERTY x = y ^ z
                """

    fg = FactorGraph(graph_desc)
    bp = BPState(fg, 2)
    assert bp.is_cyclic()


def test_cycle_detection_single_factor_with_multi():
    graph_desc = """
                NC 2
                VAR SINGLE x
                VAR SINGLE y
                VAR MULTI z
                PROPERTY x = !y
                PROPERTY x = y ^ z
                """

    fg = FactorGraph(graph_desc)
    bp = BPState(fg, 2)
    assert bp.is_cyclic()


def test_generic_single_multi():
    nc = 2
    n_exec = 2
    graph_desc = f"""
        NC {nc}

        VAR SINGLE A
        VAR SINGLE B
        VAR SINGLE C

        GENERIC SINGLE XOR

        PROPERTY XOR(A,B,C)
        """

    def xor(a, b):
        return a ^ b

    fg = FactorGraph(graph_desc)

    XOR = np.array(
        [[a, b, a ^ b] for a, b in it.product(range(nc), repeat=2)],
        dtype=np.uint32,
    )
    gen_factors = {
        "XOR": GenFactor.sparse_functional(XOR),
    }

    bp = BPState(fg, n_exec, gen_factors=gen_factors)
    bp.bp_loopy(1, True)


def test_factor_not_pub():
    nc = 4
    graph_desc = f"""
        NC {nc}

        PUB SINGLE A
        VAR SINGLE B
        PROPERTY B = !A
        """
    fg = FactorGraph(graph_desc)
    for a in range(nc):
        bp = BPState(fg, 1, public_values={"A": a})
        bp.bp_acyclic("B")
        result = bp.get_distribution("B")
        assert np.argmax(result) == (nc - 1) ^ a


def test_factor_gen_pub():
    nc = 2
    graph_desc = f"""
        NC {nc}
        PUB SINGLE A
        VAR SINGLE B
        GENERIC SINGLE NOT
        PROPERTY NOT(A,B)
        """
    fg = FactorGraph(graph_desc)
    not_factors = [
        GenFactor.sparse_functional(
            np.array([(a, (nc - 1) ^ a) for a in range(nc)], dtype=np.uint32)
        ),
        GenFactor.dense(np.array([[0, 1], [1, 0]], dtype=np.float64)),
    ]
    for nf in not_factors:
        for a in range(nc):
            bp = BPState(fg, 1, public_values={"A": a}, gen_factors={"NOT": nf})
            bp.bp_acyclic("B")
            result = bp.get_distribution("B")
            assert np.argmax(result) == (nc - 1) ^ a



import numpy as np
import numpy.typing as npt

from scalib import _scalib_ext
from scalib.config import get_config
import scalib.utils


class SNR:
    r"""Computes the Signal-to-Noise Ratio (SNR) between the traces and the
    intermediate values. Informally, SNR allows to quantify the amount of information about a
    random variable :math:`X` contained in the mean of the leakage :math:`L_X`. High SNR
    means more information contained in the mean. The SNR metric is defined
    with the following equation [1]_:

    .. math::
        \mathrm{SNR} = \frac{\mathrm{Var}_{x\leftarrow X}(\mathrm{E}[L_x])}
                {\mathrm{E}_{x\leftarrow X}(\mathrm{Var}[L_x])}

    The SNR is estimated from leakage samples :math:`L_x` and the value `x` of
    the random variable.

    Parameters
    ----------
    nc : int
        Number of possible values for the random variable :math:`X` (e.g., 256 for 8-bit
        target). `nc` must be between :math:`2` and :math:`2^{16}` (included).
    use_64bit : bool (default False)
        Use 64 bits for intermediate sums instead of 32 bits.
        When using 64-bit sums, SNR can accumulate up to :math:`2^{32}` traces, while when
        32-bit sums are used, the bound is :math:`n_i < 2^{32}/b`, where b is the
        maximum absolute value of a sample rounded to the next power of 2, and
        :math:`n_i` is the maximum number of times a variable can take a given value.
        Concretely, the total number of traces `n` should be at most
        :math:`(nc \cdot 2^{32}/b) - k `, where :math:`k = O(\sqrt{n})`, typ.
        :math:`k>=3*\sqrt{n}`  (see https://mathoverflow.net/a/273060).



    Examples
    --------
    >>> from scalib.metrics import SNR
    >>> import numpy as np
    >>> # 500 traces of 200 points, 8-bit samples
    >>> traces = np.random.randint(0,256,(500,200),dtype=np.int16)
    >>> # 10 variables on 4 bit (16 classes = 2^4)
    >>> x = np.random.randint(0,16,(500,10),dtype=np.uint16)
    >>> snr = SNR(nc=16)
    >>> snr.fit_u(traces,x)
    >>> snr_val = snr.get_snr()

    Notes
    -----
    .. [1] "Hardware Countermeasures against DPA ? A Statistical Analysis of
       Their Effectiveness", Stefan Mangard, CT-RSA 2004: 222-235

    """

    def __init__(self, nc: int, use_64bit: bool = False):
        if nc not in range(2, 2**16 + 1):
            raise ValueError(
                "SNR can be computed on max 16 bit variable (and at least 2 classes),"
                f" {nc=} given."
            )
        self._nc = nc
        self._ns = None
        self._nv = None
        self._use_64bit = use_64bit
        self._init = False

    def fit_u(self, traces: npt.NDArray[np.int16], x: npt.NDArray[np.uint16]):
        r"""Updates the SNR estimation with samples of `traces` for the classes `x`.
        This method may be called multiple times.

        Parameters
        ----------
        traces :
            Array that contains the leakage traces. The array must be of
            dimension `(n, ns)`.
        x :
            Labels for each trace. Must be of shape `(n, nv)`.
        """
        traces = scalib.utils.clean_traces(traces, self._ns)
        x = scalib.utils.clean_labels(x, self._nv)
        if not self._init:
            self._init = True
            self._ns = traces.shape[1]
            self._nv = x.shape[1]
            self._snr = _scalib_ext.SNR(self._nc, self._ns, self._nv, self._use_64bit)
        if x.shape[0] != traces.shape[0]:
            raise ValueError(
                f"Number of traces {traces.shape[0]} does not match size of classes array {x.shape[0]}."
            )
        # _scalib_ext uses inverted axes for x.
        # we can copy when needed, as x should be small, so this should be cheap
        x = np.ascontiguousarray(x.transpose())
        with scalib.utils.interruptible():
            self._snr.update(traces, x, get_config())

    def get_snr(self) -> npt.NDArray[np.float64]:
        r"""Return the current SNR estimation with an array of shape `(np,ns)`."""
        if not self._init:
            raise ValueError("Need to call .fit_u at least once.")
        with scalib.utils.interruptible():
            return self._snr.get_snr(get_config())



import numpy as np
import numpy.typing as npt
from enum import Enum, auto


class QFitMethod(Enum):
    r"""An enum class used to specify how the maximum and minimum of the traces
    is estimated based on a set of fitting traces.

    With method = QuantFitMethod.BOUNDS they are estimated as the minimum and
    maximum of the fitting trace respectively.
    With method = QuantFitMethod.MOMENT they are estimated as the average of
    the fitting traces minus/plus seven standard deviations
    """

    BOUNDS = auto()
    MOMENT = auto()


class QuantFitMethod:
    """Method for esimating the scale and shift parameters of Quantizer."""

    @classmethod
    def bounds(cls, margin=4.0):
        """Take the min and max of the training traces, fit such that the [min,
        max] range is mapped to a zero-centered interval covering a ``1/margin``
        fraction of the quantized domain: if the quantized domain is
        ``[-Q,Q]``, ``min` is mapped to ``-Q/margin`` and ``max`` is mapped to
        ``Q/margin``.
        """
        return cls(QFitMethod.BOUNDS, margin=margin)

    @classmethod
    def moment(cls, nstd=7.0):
        """Take the mean and standard deviation of the training traces, fit
        such that ``mean-nstd*std`` is mapped to `-Q` and ``mean+nstd*std`` is
        mapped to `Q`, where `[-Q, Q]` is the quantized domain.
        """
        return cls(QFitMethod.MOMENT, nstd=nstd)

    def __init__(self, method: QFitMethod, **kwargs):
        self.method = method
        self.opts = kwargs


class Quantizer:
    r"""Quantize a side channel traces given as an array of float into an array
    of int16.

    The quantizer estimates a shift and scale that minimize the loss due to the
    rounding operation.

    .. math::
        \mathrm{Quantize}( x) = \mathrm{Round}((x - \mathrm{Shift}) \cdot \mathrm{Scale})

    The shift and scale parameter can be provided explicitly, or can be
    estimated based on a few traces.

    Warning
    ^^^^^^^

    The quantization procedure operates pointwise: each point is shifted and
    scaled by a different value.
    As a consequence the quantized version of the trace probably does not look
    like its non quantized version.

    Parameters
    ----------
    shift : npt.NDArray[np.floating]
        The value to shift every traces.
    scale : npt.NDArray[np.floating]
        The value to scale every traces.

    Examples
    --------
    >>> from scalib.preprocessing import Quantizer
    >>> import numpy as np
    >>> # 500 traces of 200 points
    >>> traces : npt.NDArray[np.floating] = np.random.randn(500,200)
    >>> quantizer = Quantizer.fit(traces)
    >>> quantized_traces : npt.NDArray[np.int16] = quantizer.quantize(traces)
    >>> # Can be reused directly on 5000 new traces for instance
    >>> traces : npt.NDArray[np.floating] = np.random.randn(5000,200)
    >>> quantized_traces : npt.NDArray[np.int16] = quantizer.quantize(traces)
    """

    def __init__(
        self, shift: npt.NDArray[np.floating], scale: npt.NDArray[np.floating]
    ):
        self._shift = shift
        self._scale = scale

    @classmethod
    def fit(
        cls,
        traces: npt.NDArray[np.floating],
        method: QuantFitMethod = QuantFitMethod.bounds(),
    ):
        r"""Compute the shift and scale estimation from sample of `traces`

        This class method returns an instance of Quantizer with the
        corresponding shift and scale.

        Parameters
        ----------
        traces : array_like, np.floating
            Array that contains the traces to estimate the shift and scale in
            the quantization. The array must be of dimension `(n, ns)`
        method : QuantFitMethod
            A member of QuantFitMethod enum class that specifies how the
            minimum and maximum value of the trace to be quantized is
            estimated.
        """

        if method.method == QFitMethod.BOUNDS:
            # Max/Min Centering and Multiplication by a constant prior to
            # quantization to avoid information loss via rounding error
            max = np.amax(traces, axis=0)
            min = np.amin(traces, axis=0)
            shift = (max + min) / 2
            scale = 2**15 / ((max - min) / 2) / method.opts["margin"]
        elif method.method == QFitMethod.MOMENT:
            # Gaussian Methods
            mean = np.mean(traces, axis=0)
            std = np.std(traces, axis=0, ddof=1)
            shift = mean
            scale = 2**15 / (method.opts["nstd"] * std)
        else:
            raise ValueError("method.method should be a QFitMethod object")

        return cls(shift, scale)

    def quantize(
        self, traces: npt.NDArray[np.floating], clip: bool = False
    ) -> npt.NDArray[np.int16]:
        r"""Quantize the traces provide in `traces`

        Parameters
        ----------
        traces : array_like, np.floating
            Array that contains the traces to be quantized into int16. The array must
            be of dimension `(n, ns)`
        clip : bool
            Boolean to bypass the overflow check prior to quantization and clip
            the overflowing values to the boundaries.
            By default it is set to False.
        """
        adjusted_traces: npt.NDArray[np.floating] = (traces - self._shift) * self._scale
        if clip:
            adjusted_traces = np.clip(adjusted_traces, -(2**15), 2**15 - 1)
        else:
            overflow = (adjusted_traces > 2**15 - 1).any() or (
                adjusted_traces < -(2**15)
            ).any()
            if overflow:
                raise ValueError(
                    "Overflow detected in the quantization. Update shift and "
                    "scale more precisely to avoid the error."
                )

        quantized_traces: npt.NDArray[np.int16] = adjusted_traces.astype(np.int16)
        return quantized_traces



r"""
Soft Analytical Side-Channel Attack (SASCA)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. currentmodule:: scalib.attacks

.. autosummary::
   :toctree:
   :recursive:
   :nosignatures:

   FactorGraph
   GenFactor
   BPState
"""

__all__ = ["FactorGraph", "BPState", "GenFactor"]

from .factor_graph import FactorGraph, BPState, GenFactor



"""
Misc. internal SCALib utils.
"""

import contextlib
import signal
import threading

import numpy as np
import numpy.typing as npt


@contextlib.contextmanager
def interruptible():
    """Replace current SIGINT handler with OS-default one.

    This allows long-running non-python code (or multi-threaded one) to be
    interrupted. This results in unclean python shutdown but it is better than
    requiring to kill the process.

    This is only feasable on the main thread. In other threads, this function
    is a no-op.
    """
    if threading.current_thread() is threading.main_thread():
        restore_sig = signal.signal(signal.SIGINT, signal.SIG_DFL)
        try:
            yield
        finally:
            signal.signal(signal.SIGINT, restore_sig)
    else:
        yield


def clean_traces(traces: npt.NDArray[np.int16], ns=None):
    if not isinstance(traces, np.ndarray):
        raise ValueError(f"The traces is not a numpy array. {type(traces)}")
    elif traces.dtype != np.int16:
        raise ValueError(
            f"The traces array has dtype {traces.dtype}, expected np.int16."
        )
    elif len(traces.shape) != 2:
        raise ValueError(
            f"The traces array has {len(traces.shape)} dimensions, expected 2."
        )
    elif ns is not None and traces.shape[1] != ns:
        raise ValueError(
            f"Traces length {traces.shape[1]} does not match previously-fitted traces ({ns})."
        )
    return np.ascontiguousarray(traces)


def clean_labels(x: npt.NDArray, nv=None, multi=True, exp_type=np.uint16):
    expected_dim = 2 if multi else 1
    if not isinstance(x, np.ndarray):
        raise ValueError("The classes is not a numpy array.")
    elif x.dtype != exp_type:
        raise ValueError(f"The classes array has dtype {x.dtype}, expected {exp_type}.")
    elif len(x.shape) != expected_dim:
        raise ValueError(
            f"The classes array has {len(x.shape)} dimensions, expected {expected_dim}."
        )
    elif multi and nv is not None and x.shape[1] != nv:
        raise ValueError(
            f"Number of variables {x.shape[1]} does not match  previously-fitted classes ({nv})."
        )
    return x



import pytest
import numpy as np
from scalib.postprocessing import rank_accuracy


def test_rank_accuracy():
    nc = 256
    nsubkeys = 4
    acc = 0.2

    costs = np.zeros((nsubkeys, nc)) + 0.1
    secret_key = np.random.randint(0, nc, nsubkeys)
    costs[np.arange(nsubkeys), secret_key] = 1.0

    rmin, r, rmax = rank_accuracy(-np.log10(costs), secret_key, acc_bit=acc)

    assert r == 1.0
    assert np.log2(rmax) - np.log2(rmin) <= acc



"""General non-SCA tools that come handy when using SCALib.

Reference
---------

.. currentmodule:: scalib.tools

.. autosummary::
   :toctree:
   :recursive:
   :nosignatures:

   ContextExecutor
"""

__all__ = ["ContextExecutor"]

from concurrent.futures import ThreadPoolExecutor
import contextvars


def restore_context(context):
    for var, value in context.items():
        var.set(value)


class ContextExecutor(ThreadPoolExecutor):
    """concurrent.futures.ThreadPoolExecutor with automatic propagation of contextvars.

    The context is captured at the creation of the executor.
    """

    def __init__(self, *args, **kwargs):
        context = contextvars.copy_context()
        super().__init__(*args, **kwargs, initializer=lambda: restore_context(context))



r"""
Rank Estimation
^^^^^^^^^^^^^^^

.. currentmodule:: scalib.postprocessing

.. autosummary::
   :toctree:
   :recursive:
   :nosignatures:

   scalib.postprocessing.rankestimation
"""

__all__ = ["rankestimation"]

from .rankestimation import rank_nbin, rank_accuracy
